# Licensed under GPL version 3 - see LICENSE.rst
'''Plot routines for json output to be loaded into `three.js <threejs.org>`__.

Each routine returns a json dictionary. These json dictionaries can be collected in
a list and  written to a file with `~marxs.visualization.threesjson.write`, which
checks the formatting and adds metadata.

Just as `~marxs.visualization.threejs` this backend provides input for a webpage that
will then use the `three.js <threejs.org>`_ library to render the 3D model. The main
difference is that `~marxs.visualization.threejs` outputs plain text for each object,
while the json written by this module is much smaller if the model contains many copies
of the same object (e.g. hundreds of diffraction gratings). Also, the json data can be
updated with no changes necessary to the html page.

MARXS includes ``loader.html`` as an example how ``MARXSloader.js`` (included in MARXS
as well) can be used to load the json file and construct the three.js objects.

Note that routine to display a torus is modified
relative to the official three.js release to allow more parameters, the modified
version is included in MARXS.

For reference, here is a short summary of the json data layout.
There are two main entries: ``metadata`` (a list with the MARXS version, writer, etc...)
and ``objects`` which is a list of lists.
Each of the sublists has the following fields:

- ``n``: number of objects in list
- ``name``: string or list (if there are sevaral object in this sublist)
- ``material``: string
- ``materialpropterties``: dict
- ``geometry``: type (e.g. ``BufferGeometry``, see `three.js <threejs.org>`_)
- if ``geometry``  is a buffer Geometry: ``pos``, ``color`` (lists of numbers)
- otherwise:
  - ``pos4d``: list of lists of 16 numbers
  - ``geometrypars``: list (meaning depends on geometry, e.g. radius of torus)
'''
import os
import json
import datetime
import numpy as np
from astropy.utils.decorators import format_doc

from ..version import version
from . import threejs
from . import utils

try:
    import jsonschema
    HAS_JSONSCHEMA = True
except ImportError:
    HAS_JSONSCHEMA = False

doc_plot='''
    {__doc__}

    Parameters
    ----------
    obj : `marxs.base.MarxsElement`
        The element that should be plotted.
    display : dict of None
        Dictionary with display settings.

    Returns
    --------
    outjson : dict
        ``outjson`` is a (possibly nested)  dictionaries that describes the scene
        in form that the included MARXSloader.js can read.
'''


@format_doc(doc_plot)
def plot_object(obj, display=None, **kwargs):
    ''''Format any MARXS object as a json string.

    This method will inspect the object that is passed in and select the
    correct plotting method for its shape.
    '''
    return utils.plot_object_general(plot_registry, obj, display=display, **kwargs)


def container(obj, display=None,  **kwargs):
    ''''Recursivey output three.js json to describe all elements of a container.

    Output of each element can be a dict (if it is a leaf) or a list
    (if it is a container). We need to flatten the list here to avoid
    arbitrarily deep recursion.
    '''
    out = []
    for elem in obj.elements:
        elemout = plot_obj(elem, elem.display)
        if isinstance(elemout, list):
            out.extend(elemout)
        elif (elemout is None):
            pass
        else:
            out.append(elemout)
    return out


def box(obj, display):
    '''Describe a box-shaped optical elements.'''
    out = {}
    out['n'] = 1
    out['name'] = str(obj.name)
    out['material'] = 'MeshStandardMaterial'
    out['materialproperties'] = threejs.materialdict(display, out['material'])
    out['geometry'] = 'BoxGeometry'
    out['geometrypars'] = (2, 2, 2)
    out['pos4d'] = [obj.pos4d.T.flatten().tolist()]
    if not ('side' in display):
        out['materialproperties']['side'] = 2
    return out


def triangulation(obj, display):
    '''Describe a plane with a hole, such as an aperture of baffle.'''
    xyz, triangles = obj.geometry.triangulate(display)
    out = {}
    out['n'] = 1
    out['name'] = str(obj.name)
    out['material'] = 'MeshStandardMaterial'
    out['materialproperties'] = threejs.materialdict(display, out['material'])
    out['geometry'] = 'BufferGeometry'
    out['geometrytype'] = 'Mesh'
    out['pos'] = [xyz.flatten().tolist()]
    out['faces'] = [triangles.flatten().tolist()]

    if not ('side' in display):
        out['materialproperties']['side'] = 2

    return out


def torus(obj, display, theta0=0., thetaarc=2*np.pi, phi0=0., phiarc=np.pi * 2):
    '''Describe a (possibly incomplete) torus.

    The plot range for theta and phi is taken from the values for ``coo1`` and ``coo2``
    in the ``display`` dictionary. There entries should be a list of value if less then
    the full torus (ranging from 0 to 2 pi in each coordinate) is desired for the plot.
    '''
    theta = display.get('coo1', [0, 2 * np.pi])
    phi = display.get('coo2', [0, 2 * np.pi])

    out = {}
    out['n'] = 1
    out['name'] = str(obj.name)
    out['material'] = 'MeshStandardMaterial'
    out['materialproperties'] = threejs.materialdict(display, out['material'])
    out['geometry'] = 'ModifiedTorusBufferGeometry'
    out['geometrypars'] = (obj.R, obj.r,
                           int(np.rad2deg(theta[1])), int(np.rad2deg(phi[1])),
                           theta[1], theta[0], phi[1], phi[0])
    out['pos4d'] = [obj.pos4d.flatten().tolist()]

    if not ('side' in display):
        out['materialproperties']['side'] = 2

    return out


def plot_rays(data, scalar=None, prop={}, name='Photon list', cmap=None):
    '''Plot lines for simulated rays.

    Parameters
    ----------
    data : np.array of shape(n, N, 3)
        where n is the number of rays, N the number of positions per ray and
        the last dimension is the (x,y,z) of an Eukledian position vector.
    scalar : None or nd.array of shape (n,) or (n, N)
        This quantity is used to color the rays. If ``None`` all rays will have the same
        color. If it has n elements, each ray will have exactly one color (e.g. color
        according to the energy of the ray), if it has n*N elements, rays will be
        multicolored.
    prop : dict
        keyword arguments for line material.
    name : string
        Identifier "name" for three.js objects. This only matters if your website
        identifies elements by name for interactive features.
    cmap : `matplotlib.colors.Colormap` instance or string or None
        `matplotlib` color maps are used to convert ``scalar`` values to rgb colors.
        If ``None`` the default matplotlib colormap will be used, otherwise the colormap
        can be specified in this keyword.
    '''
    data, s_rgb, prob, n = threejs._format_plot_rays_input(data, scalar, cmap, prop)

    out = {}
    out['n'] = n
    out['name'] = name
    out['material'] = 'LineBasicMaterial'
    out['materialproperties'] = threejs.materialdict(prop, out['material'])
    out['geometry'] = 'BufferGeometry'
    out['geometrytype'] = 'Line'
    out['pos'] = data.reshape((n, -1)).tolist()
    out['color'] = s_rgb[:, :, :3].reshape((n, -1)).tolist()
    return out


def write(fileobject, data, photons=None):
    '''Add metadata and write json for three.js to disk

    Parameters
    ----------
    fileobject : writeable file-like object
    data : list of dict or single dict
        Output of ``xxx.plot(format='threejsjson')`` calls. This can either
        be a list of dictionaries or a single dictionary.
    photons : `astropy.table.Table` or None
        Some metadata is copied from a photon list, if available.
    '''
    if not isinstance(data, list):
        data = [data]
    date = datetime.datetime.now()
    jdata = {'meta': {'format_version': 1,
                      'origin': 'MARXS:threejsjson output',
                      'date': str(date.date()),
                      'time': str(date.time()),
                      'marxs_version': version},
             'elements': data}

    if photons is not None:
        data['runinfo'] = photons.meta()

    if HAS_JSONSCHEMA:
        path =  os.path.abspath(os.path.dirname(__file__))
        schemafile = os.path.join(path, 'threejs_files', 'jsonschema.json')
        with open(schemafile) as f:
            schema = json.load(f)
        jsonschema.validate(jdata, schema)

    else:
        warnings.warn('Module jsonschema not installed. json file will be written without further verification.')

    json.dump(jdata, fileobject)


plot_registry = {'triangulation': triangulation,
                 'torus': torus,
                 'box': box,
                 'container': container,
                 }



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import astropy.units as u
from ..labSource import FarLabPointSource as LabSource


def test_photon_generation():
    '''This is a test for the far source used with an aperture. It checks that
    the starting positions of the photons are within the aperture.
    '''
    pos = [1., 0., 0.]
    rate = 10 / u.s
    center = [0., 0.5, 0.]
    source = LabSource(pos, flux=rate, energy=5. * u.keV,
                       position = center, zoom = [1., 1., 2.8])

    photons = source.generate_photons(1. * u.s)
    for i in range (0, 10):
        assert photons['pos'][i][1] >= -0.5
        assert photons['pos'][i][1] <= 1.5
        assert photons['pos'][i][2] >= -2.8
        assert photons['pos'][i][2] <= 2.8



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from transforms3d.affines import decompose44

from ..math.utils import translation2aff, zoom2aff, mat2aff, e2h, h2e
from ..math.geometry import Geometry
from ..base import SimulationSequenceElement, _parse_position_keywords
import transforms3d
from transforms3d.utils import normalized_vector

__all__ = ['SimulationSetupError',
           'BaseContainer',
           'Sequence',
           'Parallel',
           'ParallelCalculated',
           'KeepCol',
           'Propagator',
           'propagate'
           ]

class SimulationSetupError(Exception):
    pass


class BaseContainer(SimulationSequenceElement):
    '''Base class for containers that contain several `SimulationSequenceElement` objects.
    '''

    display = {'shape': 'container'}
    elements = []
    '''List of elements that are contained in this container.'''

    def __init__(self, **kwargs):
        self.preprocess_steps = kwargs.pop('preprocess_steps', [])
        self.postprocess_steps = kwargs.pop('postprocess_steps', [])
        for elem in self.elements + self.preprocess_steps + self.postprocess_steps:
            if not callable(elem):
                raise SimulationSetupError('{0} is not callable.'.format(str(elem)))
        super(BaseContainer, self).__init__(**kwargs)

    def __call__(self, photons):
        for elem in self.elements:
            for p in self.preprocess_steps:
                p(photons)
            photons = elem(photons)
            for p in self.postprocess_steps:
                p(photons)
        return photons

    def elements_of_class(self, cls, subclass_ok=False, stop_at_first=False):
        '''Walk a hirachy of simulation elements to all elements of a class.

        This walks the hiracy of `marxs.simulator.Sequence` and
        `marxs.simulator.Parallel` objects. It descends depth-first.
        Typically, this function is used to
        find, e.g. all detectors of an instrument or all gratings.

        If `stop_at_first` is set it descends depth-first only until
        it find an object of type `cls`. When it does, it adds that object to
        the output list and returns to the next higher level, continuing
        the search for the remaining elements of that list. This sounds more
        complicated than it is in practice.
        Only if container types such as `marxs.simulator.Parallel` are searched
        for this becomes more complex. The search does not descend any deeper
        if a matching element is found. For example, if it finds a
        `~marxs.simulator.Parallel` object, where the elements of this object
        contain other `~marxs.simulator.Parallel` objects it will only return
        the top level one.

        Parameters
        ----------
        cls : class
            The class searched for, e.g. `marxs.optics.FlatDetector` to find all
            detectors.
        subclass_ok : bool
            Controls if subclasses of `cls` count as match.
        stop_at_first : bool
            Controls if the search continues along a branch after a match is found.

        Returns
        -------
        out : list
            Python list of all objects found.
        '''
        if subclass_ok:
            compfunc = lambda a, b: isinstance(a, b)
        else:
            compfunc = lambda a, b: type(a) == b
        a = []
        if compfunc(self, cls):
            a += [self]
            if stop_at_first:
                return a

        for e in self.elements:
            if compfunc(e, cls):
                a += [e]
            elif hasattr(e, 'elements_of_class'):
                a += e.elements_of_class(cls, subclass_ok)
        return a

class Sequence(BaseContainer):
    '''A `Sequence` is a container that summarizes several optical elements.

    Parameters
    ----------
    elements : list
        The elements of this list are all optical elements that process photons.
    preprocess_steps : list
        The elements of this list are functions or callable objects that accept a photon list as input
        and return no output (but changing the photon list in place, e.g. adding meta-data is
        allowed)  (*default*: ``[]``). All ``preprocess_steps`` are run before *every* optical element
        in the sequence.
        An example would be a function that writes the photon list to disk as a backup before
        every optical element or prints some informational message.
        If your function returns a modified photon list, treat it as an optical element and place it
        in `sequence`.
    postprocess_steps : list
        See ``preprocess_steps`` except that the steps are run *after* each sequence element
         (*default*: ``[]``).


    Examples
    --------
    The following example shows a complete marxs simulation.
    First, we import the required modules:

    >>> from astropy.coordinates import SkyCoord
    >>> import astropy.units as u
    >>> from marxs import source, optics
    >>> from marxs.simulator import Sequence

    Then, we build up the parts of the simulation, source, pointing model and hardware
    of our instrument:

    >>> mysource = source.PointSource(coords=SkyCoord(30., 30., unit="deg"), flux=1e-3 / u.s / u.cm**2, energy=2. * u.keV)
    >>> sky2mission = source.FixedPointing(coords=SkyCoord(30., 30., unit='deg'))
    >>> aper = optics.RectangleAperture(position=[50., 0., 0.])
    >>> mirr = optics.ThinLens(focallength=10, position=[10., 0., 0.])
    >>> ccd = optics.FlatDetector(pixsize=0.05)
    >>> sequence = [sky2mission, aper, mirr, ccd]
    >>> my_instrument = Sequence(elements=sequence)

    Finally, we run one set of photons through the instrument:

    >>> photons_in = mysource.generate_photons(1e5 * u.s)
    >>> photons_out = my_instrument(photons_in)

    Now, let us check where the photons fall on the detector:

    >>> set(photons_out['detpix_x'].round())
    {19.0, 20.0}

    As expected, they fall right around the center of the detector (row 19 and 20 of a
    40 * 40 pixel detector).
    '''

    def __init__(self, **kwargs):
        self.elements = kwargs.pop('elements')
        super(Sequence, self).__init__(**kwargs)


class Parallel(BaseContainer):
    '''A container for several identical optical elements.

    This object describes a set of similar elements that operate in
    parallel, meaning that each photon will only interact with one of
    them (although that is not enforced by the current
    implementation).  Examples for such an optical element would be
    the ACIS-I CCD detector on Chandra, that consists for four CCD
    chips or the HETGS gratings that are made up of many individual
    grating facets.

    The `Parallel` class requires a description of the individual
    components and a list of their positions.  This class has build-in
    support to simulate uncertainties in the manufacturing process.
    After generation, individual positions can be adjusted by hand by
    editing the ``elem_pos``.  Also, additional misalingments for each
    element can be introduced by editing ``elem_uncertainty``. This
    attribute holds a list of affine transformation matrices.  The
    global position and rotation of the combined element can be
    changed with `uncertainty`, e.g. the represent the reproducibility
    of inserting the gratings into the beam for separate observations
    or the positioning of detectors on a detector wheel. The
    uncertainty is expressed as an affine transformation matrix.

    All uncertainty matrices should only consist of translation and rotations
    and all uncertainties should be relatively small.

    After any of the attributes ``elem_pos``, ``elem_uncertainty`` or
    ``uncertainty`` is changed, `generate_elements` needs to be
    called to regenerate the positions of the individual elements using

    - the global position of ``Parallel.pos4d``
    - the position of each element ``Parallel.elem_pos`` relative to the global position
    - the global uncertainty `uncertainty`.
    - the uncertainty for individual facets.

    This mechanism can be used to estimate the influence of manufacturing
    uncertainties. First, run a simulation with optimal position, then change
    the values, regenerate the facets and rerun the simulation. Comparing the
    results will allow you to estimate the effect of the manufacturing
    misalignment.

    The order in which all the transformations are applied to the facet is
    chosen such that all rotations are done around the center of the
    individual element or the whole structure respectively.
    "Uncertainty" rotations are always done *after*
    all other rotations are accounted for.

    Parameters
    ----------
    elem_class : class
        Class of the individual elements
    elem_args : dict
        Dictionary of keyword arguments that are used to initialize
        the individual arguments. This can contain position related
        keywords as listed in `pos4d`; those will be applied to *each*
        element (e.g. set the zoom for each of them).  Usually, the
        same arguments will be applied to all facets.  In the special
        case that the value of a dictionary entry is a list with the
        same length as ``elem_pos``, one value of this list will be
        used for each element. This makes it possible to specify
        e.g. different grating constants for individual facets.  (Note
        that this has to be a list. Other python types like tuple or
        np.array that behave like lists in some contexts are not
        allowed here to avoid ambiguities.)
    elem_pos : list of arrays or dictionary of lists
        Gives the position of the individual elements in the local
        coordinate system of the `Parallel` object. This can either be
        a list of (4,4) np.arrays or a dictionary with entries of
        ``pos4d`` or ``position``, ``orientation`` and ``zoom`` as
        explained in :ref:`pos4d` where each entry in the dictionary
        is a list of values ((3,3) matrices for ``orientation``, (3,)
        vectors for ``position`` etc.).  Sub-classes of `Parallel` can
        implement a method `calculate_elempos` to determine the
        position of their elements automatically. In this case, they
        should set ``elem_pos=None``.
    id_num_offset : int
        ID numbers of generated elements start at this number.

    Examples
    --------
    In this example we build up a detector made up of four CCDs. Each
    CCD is 10 mm * 10 mm large and has a pixel size of 0.01 mm. The
    CCDs are set in a square with small spaces in between.

    >>> from marxs.simulator import Parallel
    >>> from marxs.optics import FlatDetector as CCD
    >>> detect = Parallel(elem_class=CCD, elem_args={'pixsize': 0.01, 'zoom': 5},
    ...                   elem_pos={'position': [[0, -10.1, -10.1],[0, .1, -10.1],[0, -10.1, .1],[0, .1, .1]]},
    ...                   id_col='CCD_ID')

    A column that notes which CCD was hit by each photon will be added
    to the photon table when it is processed by ``photons = detect(photons)``.
    The name of this colum will be "CCD_ID".  (If
    the `id_col` argument is not passed, the name will be the generic
    "element".)

    '''

    id_col = 'element'

    uncertainty = np.eye(4)
    '''Uncertainty of pos4d.

    The global position and rotation of the combined element can be changed with
    `uncertainty`, e.g. the represent the reproducibility of
    inserting the gratings into the beam for separate observations or the positioning
    of detectors on a detector wheel. The
    uncertainty is expressed as an affine transformation matrix.
    '''

    @property
    def pos4d(self):
        return self.geometry.pos4d

    def __init__(self, **kwargs):
        geometry = kwargs.pop('geometry', Geometry)
        if isinstance(geometry, Geometry):
            self.geometry = geometry
        elif issubclass(geometry, Geometry):
            self.geometry = geometry(kwargs)

        self.id_num_offset = kwargs.pop('id_num_offset', 0)
        self.elem_class = kwargs.pop('elem_class')
        # Need to operate on a copy here, to avoid changing elem_args of outer level
        self.elem_args = kwargs.pop('elem_args', {}).copy()

        elem_pos = kwargs.pop('elem_pos')
        if isinstance(elem_pos, dict):
            # Do some consistency checks to find the most common errors.
            keys = list(elem_pos.keys())
            for i in range(len(keys) -1):
                if ( not(hasattr(elem_pos[keys[i]], '__len__')) or
                     not(hasattr(elem_pos[keys[i + 1]], '__len__')) or
                     (len(elem_pos[keys[i]]) != len(elem_pos[keys[i + 1]]))):
                    raise ValueError('All elements in elem_pos must have the same number of entries.')

            n = len(elem_pos[keys[0]])
            # Turn dictionary of lists into list of dicts and parse position keywords
            self.elem_pos = []
            for i in range(n):
                elem_pos_dict = {}
                for k in keys:
                    elem_pos_dict[k] = elem_pos[k][i]
                self.elem_pos.append(_parse_position_keywords(elem_pos_dict))
        else:
            self.elem_pos = elem_pos

        super(Parallel, self).__init__(**kwargs)

        if 'id_col' not in self.elem_args:
            self.elem_args['id_col'] = self.id_col
        self.elem_uncertainty = [np.eye(4)] * len(self.elem_pos)
        self.generate_elements()

    def move_center(self, change4d):
        '''Move the relative origin of the coordinate system.

        Individual element positions in ``elem_pos`` are relative to the
        global ``pos4d`` of this object. This function changes the
        ``pos4d`` and adjusts the ``elem_pos`` accordingly, so that the
        elements still have the same position in global coordinates.

        This function is useful when studying uncertainties. The attribute
        ``uncertainty`` can be used to apply the same change to all elements.
        Any rotation will be applied around the reference position contained
        in ``pos4d``. Use this function to change that reference position.

        Parameters
        ----------
        change4d : np.array of shape (4,4)
            Homogeneous coordinate transformation matrix
        '''
        self.geometry.pos4d = np.dot(change4d, self.pos4d)
        inv = np.linalg.inv(change4d)
        self.elem_pos = [np.dot(inv, p) for p in self.elem_pos]
        self.generate_elements()

    def generate_elements(self):
        '''Initialize all optical elements.

        After any of the ``elem_pos``, ``elem_uncertainty`` or
        ``uncertainty`` attributes has changed, `generate_elements` needs to be
        called to regenerate the positions of the individual elements using

        - the global position of ``Parallel.pos4d``
        - the position of each element ``Parallel.elem_pos`` relative to the global position
        - the global uncertainty `uncertainty`.
        - the uncertainty for individual facets.
        '''

        self.elements = []

        for i in range(len(self.elem_pos)):
            # _parse_position_keywords pops off keywords, thus operate on a copy here
            elem_args = self.elem_args.copy()
            # check if elem_args is the same for every element
            specific_elem_args = {}
            for k, v in elem_args.items():
                if isinstance(v, list) and (len(v) == len(self.elem_pos)):
                    specific_elem_args[k] = v[i]
                else:
                    specific_elem_args[k] = v
            if 'name' not in specific_elem_args:
                specific_elem_args['name'] = 'Elem {0} in {1}'.format(i, self.name)

            elem_pos4d = _parse_position_keywords(specific_elem_args)
            telem, relem, zelem, Selem = decompose44(elem_pos4d)
            if not np.allclose(Selem, 0.):
                raise ValueError('pos4 for elem includes shear, which is not supported here.')

            e_center, e_rot, e_zoom, stemp = decompose44(self.elem_pos[i])
            tsigelem, rsigelem, zsigelem, stemp = decompose44(self.elem_uncertainty[i])
            if not np.allclose(stemp, 0.):
                raise SimulationSetupError('Shear is not supported in the elem uncertainty.')
            # Will be able to write this so much better in python 3.5,
            # but for now I don't want to nest np.dot too much so here it goes
            f_pos4d = np.eye(4)
            for m in reversed([self.pos4d,                 # global position of ParallelElement
                               self.uncertainty,           # uncertainty in global positioning
                               translation2aff(tsigelem),  # uncertaintig in translation for elem
                               translation2aff(e_center),  # translate elem center to global center
                               translation2aff(telem),     # offset for all elem. Usually 0.
                               mat2aff(rsigelem),          # uncertainty in rotation for elem
                               mat2aff(e_rot),             # Rotation of individual elem
                               mat2aff(relem),             # Rotation for all  elem, e.g. CAT gratings
                               zoom2aff(zsigelem),         # uncertainty in the zoom
                               zoom2aff(e_zoom),           # zoom of individual elem
                               zoom2aff(zelem),            # sets size for all elem
                              ]):
                assert m.shape == (4, 4)
                f_pos4d = np.dot(m, f_pos4d)
            self.elements.append(self.elem_class(pos4d = f_pos4d,
                                                 id_num=self.id_num_offset + i,
                                                 **specific_elem_args))


class ParallelCalculated(Parallel):
    '''Base class for containers that automatically determine positions of elements

    `ParallelCalculated` derives from `Parallel`. It adds a mechanism to
    automatically calculate the ``pos4d`` matrices of all elements in
    this container when the object is initialized.

    This class cannot be used directly since derived classes have to add that
    algorithm. The following text explains how that works.

    `ParallelCalculated` assumes that the ``pos4d`` matrix for each element
    can constructed from three vectors:

    - Position of the center
    - Direction of normal (x-axis of element)
    - Direction the y-axis of the element should be normal to.

    Parameters
    ----------
    pos_spec : np.array of shape (n, 4) or callable
        Specification of the center positions for all elements in homogeneous
        coordinates. If this is callable, it will be called with no argument
        and should return an np.array of shape (n, 4).
    normal_spec : np.array of shape (4, ) or callable
        Specification of the normal for each element. This can either by a point
        in homogeneous coordinates (such as the focal point of the mirror). In
        case, the normal of all elements will be directed towards that point.
        If this is a direction in homogeneous coordinates (last coordinate
        is 0), then all normal vectors are parallel to this one (e.g. to make
        CCD elements that are all perpendicular to the optical axis). Finally,
        this can be a callable. In that case, it will be called with the
        element positions from ``pos_spec`` as input and should return an array
        of the same size with normal vectors.
    parallel_spec :  np.array of shape (4, ) or callable
        Specification of a direction that one box side of each element should be
        parallel to. ``normal_spec`` defines a plane, but any rotation in that
        plane is still allowed. The element will be rotated such that one of
        the planes of the element contains the direction of ``parallel_spec``.
        If this is a homogeneous coordinate (a point or direction), see
        ``normal_spec``. If this is a callable, it will be called with two
        arguments, the output of ``pos_spec`` and ``normal_spec`` and should
        return an array of the same size.
    '''

    def __init__(self, **kwargs):
        self.pos_spec = kwargs.pop('pos_spec')
        self.normal_spec = kwargs.pop('normal_spec')
        self.parallel_spec = kwargs.pop('parallel_spec')
        kwargs['elem_pos'] = self.calculate_elempos()
        super(ParallelCalculated, self).__init__(**kwargs)

    def calculate_elempos(self):
        '''Calculate the position of elements based on some algorithm.

        Classes derived from `ParallelCalculated` can overwrite this
        method if they want to provide a way to calculate the pos4d
        matrices for the elements that make up this
        `ParallelCalculated` element.

        Returns
        -------
        pos4d : list of arrays
            List of affine transformations that bring an optical
            element centered on the origin of the coordinate system
            with the active plane in the yz-plane to the required
            facet position on the Rowland torus.

        '''
        pos4d = []

        xyzw = self.get_elemxyzw()
        normals = self.get_spec('normal_spec', xyzw)
        parallels = self.get_spec('parallel_spec', xyzw, normals)
        normals = h2e(normals)
        parallels = h2e(parallels)

        for i in range(xyzw.shape[0]):
            n = normalized_vector(normals[i, :])
            p = parallels[i, :]
            rot_mat = np.zeros((3,3))
            rot_mat[:, 0] = n
            rot_mat[:, 1] = normalized_vector(p - n * np.dot(n, p))
            rot_mat[:, 2] = normalized_vector(np.cross(n, rot_mat[:, 1]))
            pos4d.append(transforms3d.affines.compose(h2e(xyzw[i, :]), rot_mat, np.ones(3)))
        return pos4d

    def get_elemxyzw(self):
        if callable(self.pos_spec):
            return self.pos_spec()
        else:
            return self.pos_spec

    def get_spec(self, specname, xyzw, *args, **kwargs):
        spec = getattr(self, specname)
        # function
        if callable(spec):
            return spec(xyzw, *args, **kwargs)
        elif len(spec) == 4:
            # vector in homogeneous coordinates
            if spec[3] == 0:
                return np.tile(spec, (xyzw.shape[0], 1))
            # position in homogeneous coordinates
            else:
                return e2h(h2e(xyzw) - h2e(spec)[None, :], 1)
        else:
            raise ValueError(specname + 'must be callable or homogeneous coordinate.')


class KeepCol(object):
    '''Object that records the value of one column after each simulation step.

    `KeepCol` is meant to be used with the preprocess_steps or postprocess_steps
    parameters of `Sequence`. It will make a copy of one column in the photon table
    before or after each processing step and keep this copy in its `data` attribute.

    Parameters
    ----------
    colname : string
        name of column

    Attributes
    ----------
    data : list
        List of saved data. This will be empty initially.
    '''

    def __init__(self, colname):
        self.colname = colname
        self.data = []

    def __call__(self, photons):
        if self.colname not in photons.colnames:
            raise KeyError('photon list has no column {0}.'.format(self.colname))
        else:
            self.data.append(photons[self.colname].copy())

    def __array__(self):
        return np.stack(self.data)

    def format_positions(self, atol=None):
        '''Format saved position columns as a single array.

        `KeepCol` keeps the value of a column (e.g. the photon position)
        at every step of the simulation. This function reformats that list of
        columns for use graphical display programs. Note that `format_positions`
        will only work for columns that store data in homogeneous coordinates,
        for all other cases, simply call `numpy.asanyarray` on your `KeepCol`
        object.

        Parameters
        ----------
        atol : float or None
            Sometimes several consequtive elements record identical photon positions in
            `keepcol`. Those are removed from the output to speed up rendering in 3D
            programs.
            ``atol`` sets the limit up to which two positions are considered *identical*.
            See `np.allclose` for a detailed description of ``atol``.
            Set to ``None`` to skip this step.

        Returns
        -------
        pos : np.array
            Array of shape (N, n, 3), where N is the number of photons and n the number of
            unique photon positions in eukledian coordinates.
        '''
        if len(self.data) == 0:
            raise ValueError('KeepCol object contains no data.')
        if not self.data[0].ndim == 2 or self.data[0].shape[1] !=4:
            raise ValueError('format_position only works for columns that store homogeneous coordinates.')
        d = np.dstack(self.data)
        d = np.swapaxes(d, 1, 2)
        d = h2e(d)
        if atol is None:
            return d
        else:
            ind = [0]
            i = 1
            for i in range(1, d.shape[1]):
                if not np.allclose(d[:, ind[-1], :], d[:, i, :], atol=atol,
                                   equal_nan=True):
                    ind.append(i)
            return d[:, ind, :]


class Propagator(SimulationSequenceElement):
    def __init__(self, **kwargs):
        '''Move photons along the rays.

        This element moves all photons in the photon list by a specified amount
        forwards or backwards along the ray.
        Normally, photons only move forward in the simulation, but sometimes it
        can be useful to reset the photons to an earlier position, e.g. to see
        how the pattern on a detector changes for different detector positions.

        Parameters
        ----------
        distance : float
            Distance for the photons to move. Negative values move photons
            backwards.
        '''
        self.distance = kwargs.pop('distance')
        super(Propagator, self).__init__(**kwargs)

    def __call__(self, photons):
        photons['pos'] = photons['pos'] + self.distance * photons['dir']
        return photons


def propagate(photons, d):
    '''Move photons along the rays.

    This function moves all photons in the photon list by a specified amount
    forwards or backwards along the ray.
    Normally, photons only move forward in the simulation, but sometimes it
    can be useful to reset the photons to an earlier position, e.g. to see
    how the pattern on a detector changes for different detector positions.

    Parameters
    ----------
    photons : `astropy.table.Table`
        Photon table with pos and dir entries
    d : float
        Distance for the photons to move. Negative values move photons
        backwards.

    Returns
    -------
    photons : `astropy.table.Table`
        Photon table with pos and dir entries
    '''
    photons['pos'] = photons['pos'] + d * photons['dir']
    return photons



# Licensed under GPL version 3 - see LICENSE.rst
'''Gratings made by the MKI `Space Nanotechnology Laboratory`_

The MIT Kavli Institute for Astrophysics and Space Research
`Space Nanotechnology Laboratory`_ produces elements for astronomical
instruments in space. One particular field of research are critical
angle transmission (CAT) gratings, see e.g. `Heilmann et al. (2015)`_

.. _Space Nanotechnology Laboratory: http://snl.mit.edu/
.. _Heilmann et al. (2015): http://dx.doi.org/10.1117/12.2188525
'''
import numpy as np
from numpy.core.umath_tests import inner1d
from scipy.interpolate import RectBivariateSpline, interp1d
import astropy.units as u
from astropy.utils.data import get_pkg_data_filename
from astropy.table import Table

from marxs.optics import (CATGrating,
                          OrderSelector, FlatStack,
                          FlatOpticalElement)
from marxs.math.utils import norm_vector
from marxs.optics.scatter import RandomGaussianScatter


__all__ = ['l1transtab', 'l1_order_selector',
           'l1_dims', 'l2_dims',
           'qualityfactor', 'd',
           'load_table2d',
           'InterpolateEfficiencyTable',
           'QualityFactor',
           'L1',
           'L2Abs',
           'L2Diffraction',
           'CATL1L2Stack',
           'NonParallelCATGrating',
           ]

d = 0.0002
'''Spacing of grating bars'''

l1transtab = Table.read(get_pkg_data_filename('SiTransmission.csv'), format='ascii.ecsv')
'''Transmission through 1 mu of Si'''

l1_order_selector = OrderSelector(orderlist=np.array([-4, -3, -2, -1, 0, 1, 2, 3, 4]),
                                  p=np.array([0.006, 0.0135, 0.022, 0.028, 0.861, 0.028, 0.022, 0.0135, 0.006]))
'''Simple order selector for diffraction on L1.

The numbers here are calculated for the 2018 Arcus gratings assuming the L1 structure
is independent from the grating membrane itself (which is not true, but a valid first
approximation.)
'''

l1_dims = {'bardepth': 0.004 * u.mm, 'period': 0.005 * u.mm, 'barwidth': 0.0009 * u.mm}
'''Dimensions of L1 support bars running perpendicular to the grating bars'''

l2_dims = {'bardepth': 0.5 * u.mm, 'period': 0.9622504 * u.mm, 'barwidth': 0.0962250 * u.mm}
'''Dimensions of hexagonal L2 support'''

qualityfactor = {'d': 200. * u.um, 'sigma': 1.75 * u.um}
'''Scaling of grating efficiencies, parameterized as a Debye-Waller factor'''


class DataFileFormatException(Exception):
    '''Exception for grating efficiency files not matching expected format.'''
    pass


def load_table2d(filename):
    '''Get a 2d array from an ecsv input file.

    In the table file, the data is flattened to a 1d form.
    The first two columns are x and y, like this:
    The first column looks like this with many duplicates:
    [1,1,1,1,1,1,2,2,2,2,2,2,3,3,3, ...].
    Column B repeats like this: [1,2,3,4,5,6,1,2,3,4,5,6,1,2,3, ...].

    All remaining columns are data on the same x-y grid, and the grid
    has to be regular.


    Parameters
    ----------
    filename : string
        Name and path of data file

    Returns
    -------
    tab : `astropy.table.Table`
        Table as read in. Useful to access units or other meta data.
    x, y : `astropy.table.Column`
        Unique entries in first and second column
    dat : np.array
        The remaining outputs are np.arrays of shape (len(x), len(y))
    '''
    tab = Table.read(filename, format='ascii.ecsv')
    x = tab.columns[0]
    y = tab.columns[1]
    n_x = len(set(x))
    n_y = len(set(y))
    if len(x) != (n_x * n_y):
        raise DataFileFormatException('Data is not on regular grid.')

    x = x[::n_y]
    y = y[:n_y]
    coldat = [tab[d].data.reshape(n_x, n_y) for d in tab.columns[2:]]

    return tab, x, y, coldat


class InterpolateEfficiencyTable(object):
    '''Order Selector for MARXS using a specific kind of data table.

    Ralf Heilmann from the SNL typically writes simulated grating efficiencies
    into Excel tables. Since Excel is hard to read in Python and not very
    suited to version control with git, those tables are converted to csv files
    of a certain format.
    A short summary of this format is given here, to help reading the code.
    The table contains data in 3-dimenasional (wave, n_theta, order) space,
    flattened into a 2d table.

    - Row 1 + 2: Column labels. Not used here.
    - Column A: wavelength in nm.
    - Column B: blaze angle in deg.
    - Rest: data

    For each wavelength there are multiple blaze angles listed, so Column A
    contains
    many dublicates and looks like this: [1,1,1,1,1,1,2,2,2,2,2,2,3,3,3, ...].
    Column B repeats like this: [1,2,3,4,5,6,1,2,3,4,5,6,1,2,3, ...].

    Because the wave, theta grid is regular, this class can use the
    `scipy.interpolate.RectBivariateSpline` for interpolation in each 2-d slice
    (``order`` is an integer and not interpolated).

    Parameters
    ----------
    filename : string
        path and name of data file
    k : int
        Degree of spline. See `scipy.interpolate.RectBivariateSpline`.
    '''

    def __init__(self, filename, k=3):
        tab, wave, theta, orders = load_table2d(filename)
        theta = theta.to(u.rad)
        # Order is int, we will never interpolate about order,
        self.orders = np.array([int(n) for n in tab.colnames[2:]])
        self.interpolators = [RectBivariateSpline(wave, theta, d, kx=k, ky=k) for d in orders]

    def probabilities(self, energies, pol, blaze):
        '''Obtain the probabilties for photons to go into a particular order.

        This has the same parameters as the ``__call__`` method, but it returns
        the raw probabilities, while ``__call__`` will draw from these
        probabilities and assign an order and a total survival probability to
        each photon.

        Parameters
        ----------
        energies : np.array
            Energy for each photons
        pol : np.array
            Polarization for each photon (not used in this class)
        blaze : np.array
            Blaze angle for each photon

        Returns
        -------
        orders : np.array
            Array of orders
        interpprobs : np.array
            This array contains a probability array for each photon to reach a
            particular order
        '''
        # convert energy in keV to wavelength in nm
        # (nm is the unit of the input table)
        wave = (energies * u.keV).to(u.nm, equivalencies=u.spectral()).value
        interpprobs = np.empty((len(self.orders), len(energies)))
        for i, interp in enumerate(self.interpolators):
            interpprobs[i, :] = interp.ev(wave, blaze)
        return self.orders, interpprobs

    def __call__(self, energies, pol, blaze):
        orders, interpprobs = self.probabilities(energies, pol, blaze)
        totalprob = np.sum(interpprobs, axis=0)
        # Cumulative probability for orders, normalized to 1.
        cumprob = np.cumsum(interpprobs, axis=0) / totalprob
        ind_orders = np.argmax(cumprob > np.random.rand(len(energies)), axis=0)

        return orders[ind_orders], totalprob


class QualityFactor(FlatOpticalElement):
    '''Scale probabilites of theoretical curves to measured values.

    All gratings look better in theory than in practice. This grating quality
    factor scales the calculated diffraction probabilities to the observed
    performance.
    '''

    def __init__(self, qualityfactor=qualityfactor, **kwargs):
        self.factor = np.exp(- (2 * np.pi * qualityfactor['sigma'] /
                                qualityfactor['d'])**2)
        super().__init__(**kwargs)

    def specific_process_photons(self, photons, intersect, interpos, intercoos):
        return {'probability': self.factor**(photons['order'][intersect]**2)}
        return photons


def check_lx_dims(lx_dims):
    '''Check that dimensions in l1_dims or l2_dims make sense'''
    if not (lx_dims['barwidth'] < lx_dims['period']):
        raise ValueError('Period of grating must be larger than bar width.')


class L1(CATGrating):
    '''A CAT grating representing only the L1 structure

    This is treated independently of the CAT grating layer itself
    although the two gratings are not really in the far-field limit.
    CAT gratings of this class determine (statistically) if a photon
    passes through the grating bars or the L1 support.
    The L1 support is simplified as solid Si layer of 4 mu thickness.

    Parameters
    ----------
    l1_dims : dict
    '''
    blaze_name = 'blaze_L1'
    order_name = 'order_L1'

    def __init__(self, l1_dims=l1_dims, **kwargs):
        check_lx_dims(l1_dims)
        self.openfraction = 1 - l1_dims['barwidth'] / l1_dims['period']
        energy = l1transtab['energy'].to(u.keV, equivalencies=u.spectral())
        trans = np.exp(np.log(l1transtab['transmission']) * l1_dims['bardepth'] / (1 * u.micrometer))
        self.transfunc = interp1d(energy, trans)
        kwargs['d'] = l1_dims['period'].to(u.mm).value
        super().__init__(**kwargs)


    def specific_process_photons(self, photons, intersect,
                                 interpos, intercoos):
        catresult = super().specific_process_photons(photons, intersect, interpos, intercoos)

        # Now select which photons go through the L1 support and
        # set the numbers appropriately.
        # It is easier to have the diffraction calculated for all photons
        # and then re-set numbers for a small fraction here.
        # That, way, I don't have to duplicate the blaze calculation and no
        # crazy tricks are necessary to keep the indices correct.
        l1 = np.random.rand(intersect.sum()) > self.openfraction
        ind = intersect.nonzero()[0][l1]
        catresult['dir'][l1] = photons['dir'].data[ind, :]
        catresult['polarization'][l1] = photons['polarization'].data[ind, :]
        catresult['order_L1'][l1] = 0
        catresult['probability'][l1] = self.transfunc(photons['energy'][ind])
        return catresult


class L2Abs(FlatOpticalElement):
    '''L2 absorption and shadowing

    Some photons may pass through the CAT grating membrane and L1
    support, but are absorbed by the L2 sidewalls. We treat this
    statistically by reducing the overall probability.  This
    implementation ignores the effect that photons might scatter on
    the L2 sidewall surface (those would be scattered away from the
    CCDs anyway for most layouts).

    Note that this does not read the L2 from a file, but calculates it
    directly from the dimensions.

    '''
    def __init__(self, l2_dims=l2_dims, **kwargs):
        check_lx_dims(l2_dims)
        self.bardepth = l2_dims['bardepth']
        self.period = l2_dims['period']
        self.barwidth = l2_dims['barwidth']
        super().__init__(**kwargs)
        self.innerfree = self.period - self.barwidth

    def specific_process_photons(self, photons, intersect,
                                 interpos, intercoos):

        p3 = norm_vector(photons['dir'].data[intersect])
        ex, ey, en = self.geometry.get_local_euklid_bases(intercoos[intersect, :])
        angle = np.arccos(np.abs(inner1d(p3, en)))

        # fractional area NOT covered by the hexagon structure
        openfraction = (self.innerfree / self.period)**2
        # fractional area shadowed by inclined hexagon structure
        shadowarea = (self.bardepth * self.innerfree * np.sin(angle))
        totalarea = self.period**2 / 2 * np.sqrt(3)
        shadowfraction = shadowarea / totalarea

        return {'probability': openfraction - shadowfraction}


class L2Diffraction(RandomGaussianScatter):
    '''Very simple approximation of L2 diffraction effects.

    L2 is a hexagonal pattern, but at such a large spacing, that diffraction
    at higher orders can be safely neglected. The only thing that does
    happen is a slight broadening due to the single-slit function, and again,
    only the core of that matters. So, we simply approximate this here with
    simple Gaussian Scattering using the radius of the Airy disk as estimate
    for the broadening sigma.
    '''
    scattername = 'L2Diffraction'
    def __init__(self, l2_dims=l2_dims, **kwargs):
        check_lx_dims(l2_dims)
        self.innerfree = l2_dims['period'] - l2_dims['barwidth']
        super().__init__(**kwargs)

    def scatter(self, photons, intersect, interpos, intercoos):
        wave = (photons['energy'].data[intersect] * u.keV).to(u.mm, equivalencies=u.spectral())
        # 1.22 from Airy disk formula https://en.wikipedia.org/wiki/Airy_disk
        # 0.4 is approx factor between sigma and r (first minimum)
        sigma = 1.22 * 0.4 * np.arcsin(wave / self.innerfree)
        return np.random.normal(size=intersect.sum()) * sigma


class NonParallelCATGrating(CATGrating):
    '''CAT Grating where the angle of the reflective wall changes.

    This element represents a CAT grating where not all grating bar walls
    are perpendicular to the surface of the grating. This is only
    true for a ray through the center. The angle changes linearly with
    the distance to the center in the dispersion direction.
    Each grating bar has a fixed angle, i.e. no change of the direction
    happens along the grating bars (perpendicular to the dispersion direction).

    Parameters
    ----------
    d_blaze_mm : float
        Change in direction of the reflecting grating bar sidewall, which
        directly translates to a change in blaze angle [rad / mm].
    blaze_center : float
        Blaze angle at center of grating, ``0`` means grating bars are
        perpendicular to element surface. [rad]
    '''
    def __init__(self, **kwargs):
        self.d_blaze_mm = kwargs.pop('d_blaze_mm', 0)
        self.blaze_center = kwargs.pop('blaze_center', 0)
        super().__init__(**kwargs)

    def blaze_angle_modifier(self, intercoos):
        '''
        Parameters
        ----------
        intercoos : np.array
            intercoos coordinates for photons interacting with optical element
        '''
        return self.blaze_center + intercoos[:, 0] * self.d_blaze_mm


def catsupportbars(photons):
    '''Metal structure that holds grating facets will absorb all photons
    that do not pass through a grating facet.

    We might want to call this L3 support ;-)
    '''
    photons['probability'][photons['facet'] < 0] = 0.
    return photons


class CATL1L2Stack(FlatStack):
    '''SNL fabricated CAT grating

    This element combines all parts of a CAT grating into a single object.
    These include the grating membrane and the absorption and diffraction due
    to the L1 and L2 support.
    Approximations are done for all those elements, see the individial classes
    for more details. Except for `order_selector` all other parameters are set with
    defaults defined in module level variables.

    Parameters
    ----------
    order_selector : `marxs.optics.OrderSelector`
        Order selector for the grating membrane
    groove_angle : float
        Goove angle of grating bars (in rad). Default: 0
    l1_order_selector : `marxs.optics.OrderSelector`
        Order selector for L1 dispersion (cross-dispersion direction for grating)
    qualityfactor : dict
        Parameterization of grating quality scaling factor. See model level variable
        for format.
    l1_dims : dict
        Dimensions of L1 support. See module level variable for format.
    l2_dims : dict
        Dimensions of L2 support. See module level variable for format.
    '''
    def __init__(self, l1_dims=l1_dims, l2_dims=l2_dims,
                 l1_order_selector=l1_order_selector,
                 qualityfactor=qualityfactor,
                 **kwargs):
        kwargs['elements'] = [CATGrating,
                              QualityFactor,
                              L1,
                              L2Abs,
                              L2Diffraction,
                          ]
        groove_angle = kwargs.pop('groove_angle', 0.)
        kwargs['keywords'] = [{'order_selector': kwargs.pop('order_selector'),
                               'd': kwargs.pop('d', d),
                               'groove_angle': groove_angle},
                              {'qualityfactor': qualityfactor},
                              {'l1_dims': l1_dims,
                               'order_selector': l1_order_selector,
                               'groove_angle': np.pi / 2. + groove_angle},
                               {'l2_dims': l2_dims},
                               {'l2_dims': l2_dims}
                          ]
        super().__init__(**kwargs)



# Licensed under GPL version 3 - see LICENSE.rst
def get_package_data():
    # Installs the data files.  Unable to get package_data
    # to deal with a directory hierarchy of files, so just explicitly list.
    return {
        'marxs.visualization': ['threejs_files/MARXSloader.js',
                                'threejs_files/loader.html',
                                'threejs_files/ModifiedTorusBufferGeometry.js',
                                'threejs_files/jsonschema.json'
                                   ]
    }



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.table import Table
import astropy.units as u
import pytest

from ..aperture import CircleAperture, RectangleAperture, MultiAperture
from ...utils import generate_test_photons
from ...source import FixedPointing
from ...base import GeometryError
from ...math.utils import h2e

def test_circle_phi():
    p = generate_test_photons(500)
    p.remove_column('pos')
    c = CircleAperture(phi=[0, np.pi / 2])
    p = c(p)
    assert np.all(p['pos'][1] >= 0)
    assert np.all(p['pos'][2] >= 0)

def test_circle_invalid_parameters():
    with pytest.raises(ValueError) as e:
        c = CircleAperture(phi=[0, 360.])
    assert 'Did you use degrees' in str(e.value)

    with pytest.raises(ValueError) as e:
        c = CircleAperture(phi=[0.2, -0.2])
    assert 'must be greater than' in str(e.value)

    with pytest.raises(ValueError) as e:
        c = CircleAperture(phi=[-.5, 2 * np.pi])
    assert 'less than 2 pi' in str(e.value)

    with pytest.raises(ValueError) as e:
        c = CircleAperture(r_inner=0.5, zoom=.2)
    assert 'must be less than' in str(e.value)

    with pytest.raises(GeometryError) as e:
        c = CircleAperture(zoom=[2, 2, 3])
    assert 'does not have the same size' in str(e.value)

def test_circle_radius():
    '''Check that all rays actually fall into the aperture area.'''
    p = generate_test_photons(500)
    p.remove_column('pos')
    c = CircleAperture(phi=[0, np.pi / 2], r_inner=15, zoom=25)
    p = c(p)
    assert np.all(p['pos'][1] >= 0)
    assert np.all(p['pos'][2] >= 0)
    rad = np.linalg.norm(h2e(p['pos']), axis=1)
    assert np.all(rad >= 15.)
    assert np.all(rad <= 25.)

def test_circle_area():
    '''Check that the area of rings and wedges comes out correctly.'''
    c = CircleAperture(zoom=[1, 2, 2])
    assert np.isclose(c.area, np.pi * 4 * u.mm**2, atol=1e-8*u.mm**2)

    c = CircleAperture(zoom=[1, 2, 2], r_inner=1)
    assert np.isclose(c.area, np.pi * 3 * u.mm**2, atol=1e-8*u.mm**2)

    c = CircleAperture(zoom=[1, 2, 2], phi = [1.3 * np.pi, 1.8 * np.pi])
    assert np.isclose(c.area, np.pi * u.mm**2, atol=1e-8*u.mm**2)

    c = CircleAperture(zoom=[1, 2, 2], phi = [1.3 * np.pi, 1.8 * np.pi], r_inner=1)
    assert np.isclose(c.area, np.pi * 0.75 * u.mm**2, atol=1e-8*u.mm**2)


def test_MultiAperture():
    p = generate_test_photons(1000)
    p.remove_column('pos')

    a1 = CircleAperture()
    a2 = RectangleAperture(position=[0, 4, 0], zoom=2)
    a = MultiAperture(elements=[a1, a2], id_col='aper')

    p = a(p)

    # All photons go through some aperture
    assert (p['aper'] == 0).sum() + (p['aper'] == 1).sum() == 1000
    # Most photons go through the bigger aperture
    assert (p['pos'][1:] > 1).sum() > 650
    # Regression: Check apertures are not sorted in time [#189]
    ind1 = (p['aper'] == 0).nonzero()[0]
    ind2 = (p['aper'] == 1).nonzero()[0]
    assert not np.max(ind1) <  np.min(ind2)
    assert not np.min(ind1) > np.max(ind2)

def test_geomarea_projection():
    '''When a ray sees the aperture under an angle the projected aperture size
    is smaller. This is accounted for by reducing the probability of this photon.'''
    photons = Table()
    photons['ra'] = [0., 45., 90., 135., 315.]
    photons['dec'] = np.zeros(5)
    photons['origin_coord'] = SkyCoord(photons['ra'], photons['dec'], unit='deg')
    photons['probability'] = np.ones(5)
    photons['time'] = np.arange(5)
    photons['polangle'] = np.zeros(5)

    fp = FixedPointing(coords=SkyCoord(0., 0., unit='deg'))
    photons = fp(photons)
    aper = RectangleAperture()
    p = aper(photons.copy())

    assert np.allclose(p['probability'], [1., 1./np.sqrt(2), 0, 0, 1./np.sqrt(2)])

    orientation = np.array([[0, 0, 1],[1, 0, 0],[0, 1, 0]])
    aper = RectangleAperture(orientation=orientation)
    p = aper(photons.copy())

    assert np.allclose(p['probability'], [ 0, 1./np.sqrt(2), 1., 1./np.sqrt(2), 0.])



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from astropy.io import ascii

from .base import FlatOpticalElement, FlatStack
from ..math.utils import norm_vector, e2h, h2e


class FlatBrewsterMirror(FlatOpticalElement):
    '''Flat mirror operated at the Brewster angle.

    Calculation of the Fresnel coefficients can be computationally intense
    and also requires knowledge of the refractive index for the appropriate material.
    The ``FlatBrewsterMirror`` simplifies this for a mirror that is known to be
    operated at the Brewster angle.

    This mirror assumes that all photons arrive at the Brewster angle
    where only s (senkrecht = direction perpendicular to plane of incidence)
    polarisation is reflected.
    It also assumes that all photons that are not reflected (i.e. those that
    are transmitted) are lost. No transmitted photons are returned, instead the
    probability of the reflected photons is adjusted to account for this overall loss.
    '''
    display = {'color': (0., 1., 0.),
               'shape': 'box',
               'box-half': '+x',
    }

    def fresnel(self, photons, intersect, intersection, local):
        '''The incident angle can easily be calculated from e_x and photons['dir'].

        Returns
        -------
        refl_s, refl_p : np.array or float
            Reflection probability for s and p polarized photons.
            Typically, the number will depend on the incident angle and energy
            of each photon and thus the return value will be a vector.
        '''
        return 1., 0.

    def specific_process_photons(self, photons, intersect, intersection, local):
        directions = norm_vector(photons['dir'].data[intersect])
        # save the direction of the incoming photons as beam_dir
        beam_dir = h2e(directions)

        # reflect the photons (change direction) by transforming to local coordinates
        pos4d_inv = np.linalg.inv(self.pos4d)
        directions = directions.T
        directions = np.dot(pos4d_inv, directions)
        directions[0, :] *= -1
        directions = np.dot(self.pos4d, directions)
        new_beam_dir = directions.T

        # split polarization into s and p components
        # - s is s polarization (perpendicular to plane of incidence)
        # - p is p polarization (in the plane of incidence)

        # First, make basis vectors.
        v_s = np.cross(beam_dir, self.geometry['e_x'][0:3])
        v_s /= np.linalg.norm(v_s, axis=1)[:, np.newaxis]
        v_p = np.cross(beam_dir, v_s)

        polarization = h2e(photons['polarization'].data[intersect])
        p_v_s = np.einsum('ij,ij->i', polarization, v_s)
        p_v_p = np.einsum('ij,ij->i', polarization, v_p)

        fresnel_refl_s, fresnel_refl_p = self.fresnel(photons, intersect, intersection, local)
        # Calculate new intensity ~ (E_x)^2 + (E_y)^2
        Es2 = fresnel_refl_s * p_v_s ** 2
        Ep2 = fresnel_refl_p * p_v_p ** 2

        # parallel transport of polarization vector
        # v_s stays the same by definition
        new_v_p = np.cross(h2e(new_beam_dir), v_s)
        new_pol = norm_vector(-Es2[:, np.newaxis] * v_s  + Ep2[:, np.newaxis] * new_v_p)


        return {'dir': new_beam_dir,
                'probability': Es2 + Ep2,
                'polarization': e2h(new_pol, 0)}


class MultiLayerEfficiency(FlatOpticalElement):
    '''The Multilayer mirror with varying layer thickness along one axis

    The distance between layers (and thus best reflected energy) changes along
    the local y axis.
    All reflectivity data is assumed to be for a single, desired angle. There
    is currently no way to enter varying reflection that depends on the angle
    of incidence.

    Provide reflectivity data in a file with columns:

    - 'X(mm)' - position along the "changing" axis (local y axis)
    - 'Peak lambda' - wavelength with maximum reflection at a given position
    - 'Peak' - maximum reflection at a given position
    - 'FWHM(nm)' - full width half max, measure of width of reflection Gaussian peaks

    Provide polarization data in a file with columns:

    - 'Photon energy' - energy of the photon in keV
    - 'Polarization' - Fraction polarized in the more reflective direction, so that
      randomly polarized light would have a value of 0.5.

    Parameters
    ----------
    reflFile: string
        path, filename, and .txt extension for reflection data file
    testedPolarization: string
        path, filename, and .txt to a text file containing a table with photon energy
        and fraction polarization for the light used to test the mirrors and create the
        reflectivity file
    '''
    def __init__(self, **kwargs):
        self.fileName = kwargs.pop('reflFile')
        self.polFile = kwargs.pop('testedPolarization')
        super(MultiLayerEfficiency, self).__init__(**kwargs)

    def interp_files(self, photons, local):
        # read in correct reflecting probability file, now in table format
        reflectFile = ascii.read(self.fileName)

        # find reflectivity adjustment due to polarization of light in reflectivity testing
        polarizedFile = ascii.read(self.polFile)
        tested_polarized_fraction = np.interp(photons['energy'], polarizedFile['Photon energy'] / 1000, polarizedFile['Polarization'])

        # find probability of being reflected due to position
        local_x = local[:, 0] / np.linalg.norm(self.geometry['v_y'])
        local_coords_in_file = reflectFile['X(mm)'] / np.linalg.norm(self.geometry['v_y']) - 1
        # interpolate 'Peak lambda', 'Peak' [reflectivity], and 'FWHM(nm)' to the actual photon positions
        peak_wavelength = np.interp(local_x, local_coords_in_file, reflectFile['Peak lambda'])
        max_refl = np.interp(local_x, local_coords_in_file, reflectFile['Peak']) / tested_polarized_fraction
        spread_refl = np.interp(local_x, local_coords_in_file, reflectFile['FWHM(nm)'])

        return peak_wavelength, max_refl, spread_refl

    def specific_process_photons(self, photons, intersect, intersection, local):
         # wavelength is in nm assuming energy is in keV
        wavelength = 1.23984282 / photons['energy'].data[intersect]
        peak_wavelength, max_refl, spread_refl = self.interp_files(photons[intersect], local[intersect])

        # the standard deviation squared of the Gaussian reflectivity functions of each photon's wavelength
        c_squared = (spread_refl ** 2) / (8. * np.log(2))
        # skip the case when there is no Gaussian (this is assumed to just be the zero function)
        c_is_zero = (c_squared == 0)

        refl_prob = np.zeros(len(wavelength))
        refl_prob[~c_is_zero] = max_refl[~c_is_zero] * np.exp(-((wavelength[~c_is_zero] - peak_wavelength[~c_is_zero]) ** 2) / (2 * c_squared[~c_is_zero]))
        return {'probability': refl_prob / 100}

class MultiLayerMirror(FlatStack):
    def __init__(self, **kwargs):
        super(MultiLayerMirror, self).__init__(elements=[FlatBrewsterMirror, MultiLayerEfficiency],
                                               keywords=[{}, {'reflFile': kwargs.pop('reflFile'),
                                                              'testedPolarization': kwargs.pop('testedPolarization')}],
                                               **kwargs)



"""This module make plots and sketches for the documentation."""

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import numpy as np

def optical_element_coords():
   fig = plt.figure()
   ax = fig.add_subplot(111, projection='3d')
   ax.quiver([2,0,0],[0,2,0],[0,0, 2],[1,0,0],[0,1,0], [0,0,1], length=4, arrow_length_ratio=.05, color='k', lw=2)
   ax.bar3d(-1, -1, -1, 1, 2, 2, zsort='average', alpha=0.5, color=['b'] *5 + ['r'], zorder=1)
   ax.set_axis_off()
   ax.text(2.1, 0, 0, 'x', None, fontsize='x-large')
   ax.text(0, 2.1, 0, 'y', None, fontsize='x-large')
   ax.text(0, 0, 2.1, 'z', None, fontsize='x-large')
   ax.set_xlim([-1.2,1.2])
   ax.set_ylim([-1.2,1.2])
   ax.set_zlim([-1.2,1.2])
   fig.subplots_adjust(left=0,right=1,top=1,bottom=0)
   plt.show()



# Licensed under GPL version 3 - see LICENSE.rst
'''Geometry and operations of the Chandra Observatory.

This module provides the optical elements on the Chandra Observatory.

.. warning:: Currently incomplete.

   Missing:

    - a lot

   Incomplete:

   - Grating order efficiencies are the same for every order.

Some spacecraft properties are defined in module attributes, others are read from
reference files in the CALDB (or the files shipped with `classic MARX`_).
Properties that are very easy to define and that are very unlikely to ever change
(e.g. the mapping of ACIS chip number to chip name) are set in this module.
That makes is easier to see where the numbers come from and thus helps for users
who look at this module as an example of how to build up complex marxs setups.
'''
from .hess import HETG
from .chandra import Chandra
from .dither import LissajousDither
from .acis import ACIS
from .data import (NOMINAL_FOCALLENGTH, AIMPOINTS, PIXSIZE)
from .hrma_py import Aperture, HRMA



# Licensed under GPL version 3 - see LICENSE.rst
from .aperture import RectangleAperture, CircleAperture, MultiAperture
from .detector import FlatDetector, CircularDetector
from .marx import MarxMirror
from .grating import (FlatGrating, CATGrating,
                      OrderSelector, EfficiencyFile
                      )
from .mirror import ThinLens, PerfectLens
from .baffle import Baffle
from .scatter import RadialMirrorScatter, RandomGaussianScatter
from .filter import EnergyFilter, GlobalEnergyFilter
from .base import OpticalElement, FlatOpticalElement, FlatStack
from .multiLayerMirror import FlatBrewsterMirror



# Licensed under GPL version 3 - see LICENSE.rst
from .base import (GeometryError,
                  DocMeta,
                  MarxsElement, SimulationSequenceElement,
                  _parse_position_keywords
                  )



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from scipy.stats import kstest
import transforms3d
import pytest
from astropy.coordinates import SkyCoord
import astropy.units as u

from ..rowland import (RowlandTorus, GratingArrayStructure, LinearCCDArray,
                       RowlandCircleArray,
                       find_radius_of_photon_shell, design_tilted_torus,
                       ElementPlacementError)
from ...optics.base import FlatOpticalElement
from ...source import PointSource, FixedPointing
from ...optics import MarxMirror, OrderSelector, FlatGrating
from ...math.utils import h2e

class mock_facet(FlatOpticalElement):
    '''Lightweight class with no functionality for tests.'''
    pass

def test_radius_of_photon_shell():
    mysource = PointSource(coords=SkyCoord(30., 30., unit="deg"))
    photons = mysource.generate_photons(1. * u.ks)
    mypointing = FixedPointing(coords=SkyCoord(30, 30., unit='deg'))
    photons = mypointing.process_photons(photons)
    marxm = MarxMirror('./marxs/optics/hrma.par', position=np.array([0., 0, 0]))
    photons = marxm(photons)
    r1, r2 = find_radius_of_photon_shell(photons, 1, 9e3)
    assert abs(r1 - 433.) < 1.
    assert abs(r2 - 442.) < 1.
    r1, r2 = find_radius_of_photon_shell(photons, 1, 9e3, percentile=[49, 49.001])
    assert (r2 - r1) < 0.01

def test_design_tilted_torus():
    '''Test the trivial case with analytic answers and consistency for other angles'''
    R, r, pos4d = design_tilted_torus(10, 0., 0.)
    assert R == 5
    assert r == 5
    assert np.all(pos4d == np.eye(4))
    # This is a solution that looks good by hand, but could still be slightly wrong...
    R, r, pos4d = design_tilted_torus(10, np.deg2rad(3), np.deg2rad(6))
    assert r == 5.0068617299896054
    assert R == 4.9794336175561327
    assert pos4d[0,3] == 0.027390523158633273
    # The z axis should be unchanged
    assert np.allclose(pos4d[[0,1],2], 0)

def test_design_tilted_torus_negative_angles():
    '''Tilt the torus the other way. Should be same geometry in general.'''
    R, r, pos4d = design_tilted_torus(10, np.deg2rad(3), np.deg2rad(6))
    Rn, rn, pos4dn = design_tilted_torus(10, np.deg2rad(-3), np.deg2rad(-6))
    assert R == Rn
    assert r == rn
    assert pos4d[0,3] == pos4dn[0, 3]
    assert pos4d[2, 3] == 0 # torus in xy plane
    assert pos4dn[2, 3] == 0
    assert pos4d[1, 3] == - pos4dn[1, 3]

def test_torus():
    '''Compare parametric equation of torus with non-parametric.
    '''
    R = 2.
    r = 1.

    angle = np.arange(0, 2 * np.pi, 0.1)
    phi, theta = np.meshgrid(angle, angle)
    phi = phi.flatten()
    theta = theta.flatten()
    mytorus = RowlandTorus(R=R, r=r)

    xyzw = mytorus.parametric(theta, phi)
    xyz = h2e(xyzw)

    assert np.allclose(mytorus.quartic(xyz), 0.)
    # This torus has pos4d=eye(4), so the same results should come out with this shortcut
    assert np.allclose(mytorus.quartic(xyz, transform=False), 0.)

def test_torus_solve_quartic():
    '''solve_quartic helps to find point on the torus if two coordinates are fixed.'''
    mytorus = RowlandTorus(2., 1.)
    xyzw = mytorus.parametric([.34, 1.23], [.4, 4.5])
    xyz = h2e(xyzw)

    for i in [0, 1]:
        for j in range(3):
            kwargs = {'x': xyz[i, 0], 'y': xyz[i, 1], 'z': xyz[i, 2],
                      'interval' : np.array([-0.1, 0.1]) + xyz[i, j]}
            kwargs['xyz'[j]] = None
            assert np.allclose(xyz[i, j], mytorus.solve_quartic(**kwargs))

    # Nothing to solve for
    with pytest.raises(ValueError) as e:
        out = mytorus.solve_quartic(x=1, y=1, z=1)
    assert 'Exactly one of the input numbers' in str(e.value)

    # Too many parameters to solve for
    with pytest.raises(ValueError) as e:
        out = mytorus.solve_quartic(x=1, y=None, z=None)
    assert 'Exactly one of the input numbers' in str(e.value)

def test_rotated_torus():
    '''Test the torus equation for a set of points.

    Importantly, we use a parametric description of the torus here, which is
    different from the actual implementation.
    '''
    def parametrictorus(R, r, theta, phi):
        '''Just another way to specify a torus with y-axis as symmetry'''
        x = (R + r * np.cos(theta)) * np.cos(phi)
        z = (R + r * np.cos(theta)) * np.sin(phi)
        y = r * np.sin(theta)
        return np.array([x, y, z]).T

    angle = np.arange(0, 2 * np.pi, 0.1)
    phi, theta = np.meshgrid(angle, angle)
    phi = phi.flatten()
    theta = theta.flatten()
    xyz = parametrictorus(2., 1., theta, phi)
    # rotate -90 deg around y axis:
    #   x -> z
    #   z -> -x
    rotatedxyz = np.zeros_like(xyz)
    rotatedxyz[:, 0] = -xyz[:, 2]
    rotatedxyz[:, 1] = xyz[:, 1]
    rotatedxyz[:, 2] = xyz[:, 0]
    mytorus = RowlandTorus(R=2., r=1., orientation=transforms3d.axangles.axangle2mat([0,1,0], -np.pi/2))

    assert np.allclose(mytorus.quartic(rotatedxyz), 0.)

    assert np.allclose(rotatedxyz, h2e(mytorus.parametric(theta, phi)))


def test_torus_xyzw2parametric():
    '''Theta, phi coordinates for torus should round-trip to xyz'''
    R = 2.
    r = 1.

    angle = np.arange(0, 2 * np.pi, 0.1)
    phi, theta = np.meshgrid(angle, angle)
    phi = phi.flatten()
    theta = theta.flatten()
    mytorus = RowlandTorus(R=R, r=r)

    xyzw = mytorus.parametric(theta, phi)
    t, p = mytorus.xyzw2parametric(xyzw)
    assert np.allclose(np.mod(theta, 2. * np.pi), np.mod(t, 2. * np.pi))
    assert np.allclose(np.mod(phi, 2. * np.pi), np.mod(p, 2. * np.pi))


def test_torus_normal():
    '''Compare the analytic normal with a numeric result.

    The test computes a vector along the torus for small d_theta
    and d_phi. Both of those should be perpendicular to the normal.
    '''
    R = 2.
    r = 1.

    angle = np.arange(0, 2 * np.pi, 0.1)
    phi, theta = np.meshgrid(angle, angle)
    phi = phi.flatten()
    theta = theta.flatten()
    mytorus = RowlandTorus(R=R, r=r)

    xyz = mytorus.parametric(theta, phi)
    t1 = mytorus.parametric(theta + 0.00001, phi)
    p1 = mytorus.parametric(theta, phi + 0.00001)
    t0 = mytorus.parametric(theta - 0.00001, phi)
    p0 = mytorus.parametric(theta, phi - 0.00001)

    vec_normal = h2e(mytorus.normal(xyz))

    vec_delta_theta = h2e(t1) - h2e(t0)
    vec_delta_theta = vec_delta_theta  / np.linalg.norm(vec_delta_theta, axis=1)[:, None]
    vec_delta_phi = h2e(p1) - h2e(p0)
    vec_delta_phi = vec_delta_phi  / np.linalg.norm(vec_delta_phi, axis=1)[:, None]

    assert np.allclose(np.sqrt(np.sum(vec_normal * vec_normal, axis=1)), 1.)
    assert np.allclose(np.einsum('ij,ij->i', vec_normal, vec_delta_theta), 0.)
    assert np.allclose(np.einsum('ij,ij->i', vec_normal, vec_delta_phi), 0.)

def torus_cut_is_circle():
    '''Cutting the torus in a plane that includes the symmetry axis should give a cirlce.

    Check that that's true even for a tilted torus.
    '''
    R, r, pos4d = design_tilted_torus(10, np.deg2rad(-3), np.deg2rad(-6))
    torus = RowlandTorus(R, r, pos4d=pos4d)

    z = np.arange(0., torus.r *0.8, 5)
    x = np.zeros_like(z)
    xguess = torus.r + torus.R
    for i, iz in enumerate(z):
        x[i] = torus.solve_quartic(None, 0, iz, [0.99 * xguess, 1.1 * xguess])
        xguess = x[i]

    # This should be a circle of radius r around a center.
    # Where exactly is the center?
    # Assuming no y rotation:
    circle_center = np.dot(torus.pos4d[:3,:3], torus.R * np.array([1, 0,0])) + torus.pos4d[:3,3]
    circle_center_xz = circle_center[[0,2]]
    p = np.vstack([x,z])
    d = np.linalg.norm(p - circle_center_xz[:, np.newaxis], axis=0)
    assert np.allclose(d, torus.r)

def test_GratingArrayStructure():
    '''Check a GAS for consistency

    The expected numbers for this one are hand-checked and plotted and thus
    can be taken as known-good to prevent regressions later.
    '''
    myrowland = RowlandTorus(10000., 10000.)
    gas = GratingArrayStructure(myrowland, 30., [10000., 20000.], [300., 600.], phi=[-0.2*np.pi, 0.2*np.pi], elem_class=mock_facet)
    assert gas.max_elements_on_arc(300.) == 12
    angles = gas.distribute_elements_on_arc(315.) % (2. * np.pi)
    # This is a wrap-around case. Hard to test in general, but here I know the numbers
    assert np.alltrue((angles < 0.2 * np.pi) | (angles > 1.8 * np.pi))
    assert gas.max_elements_on_radius(gas.radius) == 10
    assert np.all(gas.distribute_elements_on_radius() == np.arange(315., 600., 30.))
    assert len(gas.elem_pos) == 177
    center = gas.calc_ideal_center()
    assert center[2] == 0
    assert center[1] == (300. + 600.) / 2.
    assert 2e4 - center[0] < 20.  # R_rowland >> r_gas  -> center close to R_rowland
    # fig, axes = plt.subplots(2,2)
    # for elem in [[axes[0, 0], 0, 1], [axes[0, 1], 0, 2], [axes[1, 0], 1, 2]]:
    #     for f in gas.elem_pos:
    #         elem[0].plot(f[elem[1],3], f[elem[2],3], 's')
    # plt.show()  # or move mouse in window or something.

    # There is a test missing here that the rotational part works correctly.
    # I just cannot think of a good way to check that right now.

    # Check that initially all uncertainties are 0
    for i in range(len(gas.elem_pos)):
        assert np.allclose(gas.elem_pos[i], gas.elements[i].pos4d)

def test_GAS_multipleradii():
    '''Radius can be a list of many pairs.'''
    myrowland = RowlandTorus(1000., 1000.)
    gas1 = GratingArrayStructure(myrowland, 60., [1000., 2000.], [300., 400.], elem_class=mock_facet)
    gas2 = GratingArrayStructure(myrowland, 60., [1000., 2000.], [500., 540.], elem_class=mock_facet)
    gas = GratingArrayStructure(myrowland, 60., [1000., 2000.], [300., 400., 500., 540.], elem_class=mock_facet)

    r1 = gas1.distribute_elements_on_radius()
    r2 = gas2.distribute_elements_on_radius()
    r = gas.distribute_elements_on_radius()
    assert np.all(np.in1d(r, np.union1d(r1, r2)))


def test_GratingArrayStructure_2pi():
    '''test that delta_phi = 2 pi means "full circle" and not 0
    '''
    myrowland = RowlandTorus(10000., 10000.)
    gas = GratingArrayStructure(myrowland, 30., [10000., 20000.], [300., 600.], phi=[0, 2*np.pi], elem_class=mock_facet)
    assert gas.max_elements_on_arc(300.) > 10
    n = len(gas.elem_pos)
    yz = np.empty((n, 2))
    for i, p in enumerate(gas.elem_pos):
        yz[i, :] = p[1:3, 3]
    phi = np.arctan2(yz[:, 1], yz[:, 0])
    ks, pvalue = kstest((phi + np.pi) / (2 * np.pi), 'uniform')
    assert pvalue > 0.3  # It's not exactly uniform because of finite size of elements.

def test_GAS_facets_on_radius():
    '''test distribution of elements on radius for d_r is non-integer multiple of d_element.'''
    myrowland = RowlandTorus(1000., 1000.)
    gas = GratingArrayStructure(myrowland, 60., [1000., 2000.], [300., 400.], elem_class=mock_facet)
    assert np.all(gas.distribute_elements_on_radius() == [320., 380.])
    gas.radius = [300., 340.]
    assert gas.distribute_elements_on_radius() == [320.]

def test_facet_rotation_via_facetargs():
    '''The numbers for the blaze are not realistic.'''
    gratingeff = OrderSelector(np.arange(-3, 4))
    mytorus = RowlandTorus(9e3/2, 9e3/2)
    mygas = GratingArrayStructure(mytorus, d_element=60., x_range=[5e3,1e4], radius=[538., 550.], elem_class=FlatGrating, elem_args={'zoom': 30, 'd':0.0002, 'order_selector': gratingeff})
    blaze = transforms3d.axangles.axangle2mat(np.array([0,1,0]), np.deg2rad(15.))
    mygascat = GratingArrayStructure(mytorus, d_element=60., x_range=[5e3,1e4], radius=[538., 550.], elem_class=FlatGrating, elem_args={'zoom': 30, 'orientation': blaze, 'd':0.0002, 'order_selector': gratingeff})
    assert np.allclose(np.rad2deg(np.arccos(np.dot(mygas.elements[0].geometry['e_x'][:3], mygascat.elements[0].geometry['e_x'][:3]))), 15.)

def test_persistent_facetargs():
    '''Make sure that facet_args is still intact after generating facets.

    This is important to allow tweaks of a single parameter and then to regenerate the facets.
    '''
    gratingeff = OrderSelector(np.arange(-3, 4))
    mytorus = RowlandTorus(9e4/2, 9e4/2)
    # id_col is automatically added in GAS is not present here.
    # So, pass in an id_col to make sure the comparison below will still work.
    facet_args = {'zoom': 30, 'd':0.0002, 'order_selector': gratingeff, 'id_col': 'facet'}
    mygas = GratingArrayStructure(mytorus, d_element=60., x_range=[5e4,1e5], radius=[5380., 5500.], elem_class=FlatGrating, elem_args=facet_args)
    assert mygas.elem_args == facet_args

def test_run_photons_through_gas():
    '''And check that they have the expected labels.

    No need to check here that the grating equation works - that's part of the grating tests/
    '''
    # Setup only.
    mysource = PointSource(coords=SkyCoord(30., 30., unit="deg"))
    photons = mysource.generate_photons(1. * u.ks)
    mypointing = FixedPointing(coords=SkyCoord(30, 30., unit='deg'))
    photons = mypointing.process_photons(photons)
    marxm = MarxMirror('./marxs/optics/hrma.par', position=np.array([0., 0, 0]))
    photons = marxm(photons)
    gratingeff = OrderSelector([-1, 0, 1])
    facet_args = {'zoom': 30, 'd':0.0002, 'order_selector': gratingeff}
    mytorus = RowlandTorus(9e3/2, 9e3/2)

    # Now the real test with different input for facet_args

    for i in [1, 2, 3, 4]:
        if i == 1:
            kwargs = {}
            f = 'facet'
        elif i == 2:
            kwargs = {'id_col': 'qewr'}
            f = 'qewr'
        elif i == 3:
            facet_args['id_col'] = 'ttt'
            f = 'ttt'
            kwargs = {}
        elif i == 4:
            facet_args['id_col'] = 'uuu'
            f = 'uuu'
            kwargs = {'id_col': 'yyy'}

        mygas = GratingArrayStructure(mytorus, d_element=60., x_range=[5e3,1e4], radius=[538., 550.], elem_class=FlatGrating, elem_args=facet_args, **kwargs)

        p = mygas(photons.copy())
        indorder = np.isfinite(p['order'])
        indfacet = p[f] >=0
        assert np.all(indorder == indfacet)

        assert set(p['order'][indorder]) == set([-1, 0, 1])
        # -1 means no hit - passing between facets
        allfacets = set([-1]).union(set(np.arange(len(mygas.elements))))
        assert set(p[f]).issubset(allfacets)

def test_LinearCCDArray():
    '''Test an array in default position'''
    myrowland = RowlandTorus(10000., 10000.)
    # Along the way we normally would orient the detector.
    ccds = LinearCCDArray(myrowland, d_element=30., x_range=[0., 2000.],
                          radius=[-100., 100.], phi=0., elem_class=mock_facet)
    assert len(ccds.elements) == 7
    for e in ccds.elements:
        # For this test we don't care if z and e_z are parallel or antiparallel
        assert np.isclose(np.abs(np.dot([0, 0, 1, 0], e.geometry['e_z'])), 1.)
        assert (e.pos4d[0, 3] >= 0) and (e.pos4d[0, 3] < 1.)

    # center ccd is on the optical axis
    assert np.allclose(ccds.elements[3].geometry['e_y'], [0, 1, 0, 0])

def test_LinearCCDArray_rotatated():
    '''Test an array with different rotations.

    In this case, we rotate the Rowland torus by -30 deg and then
    check that the e_z vector is rotated 30 deg with respect to the
    coordinate system.
    '''
    pos4d = transforms3d.axangles.axangle2aff([1, 0, 0], np.deg2rad(-30))
    myrowland = RowlandTorus(10000., 10000., pos4d=pos4d)
    # Along the way we normally would orient the detector.
    ccds = LinearCCDArray(myrowland, d_element=30., x_range=[0., 2000.],
                          radius=[-100., 100.], phi=0., elem_class=mock_facet)
    assert len(ccds.elements) == 7
    for e in ccds.elements:
        assert np.isclose(np.dot([0, -0.8660254, 0.5], e.geometry['e_z'][:3]), 0, atol=1e-4)

def test_impossible_LinearCCDArray():
    '''The rotation is chosen such that all requested detector positions are
    INSIDE the rowland torus
    '''
    pos4d = transforms3d.axangles.axangle2aff([1, 0, 0], np.deg2rad(-30))
    myrowland = RowlandTorus(10000., 10000., pos4d=pos4d)
    with pytest.raises(ElementPlacementError) as e:
        ccds = LinearCCDArray(myrowland, d_element=30., x_range=[0., 2000.],
                              radius=[-100., 100.], phi=np.deg2rad(30.), elem_class=mock_facet)
    assert 'No intersection with Rowland' in str(e.value)

def test_RowlandCircleArrayone():
    '''Test an array of one element in default position'''
    myrowland = RowlandTorus(10000., 10000.)
    # Along the way we normally would orient the detector.
    ccds = RowlandCircleArray(myrowland, d_element=300., theta=[np.pi - 0.0001, np.pi + 0.0001],
                              elem_class=mock_facet)
    assert len(ccds.elements) == 1
    # multiply because of pointing inward vs. outward
    assert np.allclose(ccds.elements[0].pos4d, np.eye(4) * np.array([-1,-1,1,1]))

def test_compareRowlandCircle2LinearCCD_rotatated():
    '''Test an array with different rotations.

    The RowlandCircleArray and the LinearCCDArray should be very similar
    for phi = 0.
    '''
    myrowland = RowlandTorus(10000., 10000.)
    ccd1 = RowlandCircleArray(myrowland, d_element=30., theta=[np.pi - 0.03, np.pi],
                              elem_class=mock_facet)
    ccd2 = LinearCCDArray(myrowland, d_element=30., x_range=[0., 2000.],
                          radius=[0., 10000. * np.sin(0.03)], phi=0., elem_class=mock_facet)
    # reversed because the one element orders it differently then the other.
    # Should really compare sets here or standardize order in some way, I guess.
    for e1, e2 in zip(ccd1.elements, reversed(ccd2.elements)):
        t1, r1, z1, s1 = transforms3d.affines.decompose44(e1.pos4d)
        t2, r2, z2, s2 = transforms3d.affines.decompose44(e2.pos4d)
        assert np.allclose(t1, t2, rtol=1e-3, atol=0.01)
        # orientation is not too important, so allow some -1 factors
        assert np.allclose(np.abs(r1), np.abs(r2), rtol=1e-3, atol=0.01)

def test_nojumpsinCCDorientation():
    '''Regression test: CCD orientation should change smoothly along the circle.'''
    R, r, pos4d = design_tilted_torus(12e3, 0.07, 0.15)
    rowland = RowlandTorus(R, r, pos4d=pos4d)
    det = RowlandCircleArray(rowland=rowland,
                             elem_class=mock_facet,
                             d_element=49.652, #theta=[np.pi - 0.2, np.pi + 0.5])
                             theta=[np.pi - 0.2, np.pi - 0.1])

    # If all is right, this will vary very smoothly along the circle.
    xcomponent = np.array([e.geometry['e_y'][0] for e in det.elements])
    xcompdiff = np.diff(xcomponent)
    assert (np.max(np.abs(xcompdiff)) / np.median(np.abs(xcompdiff))) < 1.1



# Autogenerated by Astropy-affiliated package marxs's setup.py on 2020-11-06 18:53:41 UTC
from __future__ import unicode_literals
import datetime

version = "1.2"
githash = "47f3960729640c8609aee1e4c80b552cf598248d"


major = 1
minor = 2
bugfix = 0

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2020, 11, 6, 18, 53, 41)
debug = False

astropy_helpers_version = "1.2.dev1156"



import numpy as np
from ...utils import generate_test_photons
from ..analysis import find_best_detector_position, mean_width_2d
from ...simulator import propagate

def test_best_detector_position():
    photons = generate_test_photons(100)
    photons['dir'].data[:, 1] = np.random.rand(100)
    photons['pos'].data[:, 1] = 0.5
    photons = propagate(photons, -10)
    # det_x is along the second axis
    out = find_best_detector_position(photons,
                                      objective_func_args={'colname': 'det_x'})
    assert np.isclose(out.x, 1.)


def test_best_detector_position_rotated():
    '''Now use z axis as optical axis'''
    photons = generate_test_photons(100)
    photons['dir'].data[:, 0] = np.random.rand(100)
    photons['dir'].data[:, 2] = - 1
    photons['pos'].data[:, 2] = 0.5
    photons = propagate(photons, -10)
    # det_x is along the second axis
    out = find_best_detector_position(photons,
                                      objective_func_args={'colname': 'det_x'},
                                      orientation=np.array([[ 0.,  1.,  0.],
                                                            [ 0.,  0.,  1.],
                                                            [ 1.,  0.,  0.]]))
    assert np.isclose(out.x, 0.5)


def test_best_detector_position_2d():
    '''Optimize a 2d distribution on the detector'''
    photons = generate_test_photons(100)
    photons['dir'].data[:, 2] = np.random.rand(100)
    photons['pos'].data[:, 1] = 0.5
    photons = propagate(photons, -10)
    # det_x is along the second axis
    out = find_best_detector_position(photons, mean_width_2d, {})
    assert np.isclose(out.x, 1.0)


# from ..gratings import resolvingpower_per_order

# def test_robust_resolvingpower():
#     '''
#         - Pass in an instance of an optical element (e.g. a
#           `marxs.optics.FlatDetector`).
#         - Pass in a `marxs.design.RowlandTorus`. This function will generate
#           a detector that follows the Rowland circle.
#         - ``None``. A flat detector in the yz plane is used, but the x position
#           for this detector is numerically optimized in each step.
#     '''
#     resolvingpower_per_order(gratings, photons, orders, detector=None,
#                              colname='det_x')



'''A pure-python approximation of the Chandra mirror.

This approximation is crude, but it's useful to have for running
examples in environments where  `classic MARX`_ is not installed.
'''
import numpy as np
import astropy.units as u

from .data import NOMINAL_FOCALLENGTH
from ...import optics

mirror_radii = np.array([[598.,610.],[481, 491], [424, 433], [315,322]])

class Aperture(optics.MultiAperture):
    '''Chandra opening aperture of four rings above the mirror shells.
    '''
    def __init__(self, **kwargs):
        if not 'id_col' in kwargs:
            kwargs['id_col'] = 'mirror_shell'
        if not 'elements' in kwargs:
            kwargs['elements'] = [optics.CircleAperture(position=[NOMINAL_FOCALLENGTH, 0, 0],
                              zoom=[1, r[1], r[1]], r_inner=r[0]) for r in mirror_radii]
        super(Aperture, self).__init__(**kwargs)


class HRMA(optics.FlatStack):
    '''High-resolution mirror assembly in a pure Python implementation

    This class delivers a rough approximation (a few percent for on-axis sources)
    to the Chandra mirror. Many important effects are missing (e.g. the mirror
    reflectivity in this module is not energy dependent, neither is the mirror scatter).
    Use `marxs.optics.MarxMirror` if possible.
    '''
    def __init__(self, **kwargs):
        'asf'
        kwargs['position'] = [NOMINAL_FOCALLENGTH, 0, 0]
        kwargs['zoom'] = [1, 650., 650]
        kwargs['elements'] = [optics.PerfectLens,
                              optics.RadialMirrorScatter,
                              optics.EnergyFilter]
        kwargs['keywords'] = [{'focallength': NOMINAL_FOCALLENGTH},
                              {'inplanescatter': 3.6e-6 * u.rad,
                               'perpplanescatter': 1.2e-6 * u.rad},
                              {'filterfunc': lambda x: np.ones_like(x) * 0.66,
                               'name': 'support spider, reflectivity, et al.'}]
        super(HRMA, self).__init__(**kwargs)



import numpy as np
from astropy.table import Table
import warnings

__all__ = ['SimulationSetupWarning', 'generate_test_photons']


class SimulationSetupWarning(Warning):
    pass

warnings.filterwarnings("always", ".*", SimulationSetupWarning)


def generate_test_photons(n=1):
    '''Generate a photon structure for testing.

    This function returns a table of ``n`` identical photons with
    standard properties:

    - position: ``x=1; y=0; z=0``
    - direction: negative x-axis
    - energy: 1 keV
    - polarization: y-direction

    This is useful for testing purposes.

    Parameters
    ----------
    n : int
        number of photons to generate

    Returns
    -------
    photons : `astropy.table.Table`
        Table of ``n`` identical photons.
    '''
    dir = np.tile(np.array([-1., 0., 0., 0.]), (n, 1))
    pos = np.tile(np.array([1., 0., 0., 1.]), (n, 1))
    pol = np.tile(np.array([0., 1., 0., 0.]), (n, 1))
    photons = Table({'pos': pos,
                     'dir': dir,
                     'energy': np.ones(n),
                     'polarization': pol,
                     'probability': np.ones(n),
                     })
    return photons



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np

from .. import HETG

def test_orient():
    '''Check orientation of gratings.'''
    hetg = HETG()
    for i in range(len(hetg.elements)):
        ex = hetg.elements[i].geometry['e_x'][:3]
        ey = hetg.elements[i].geometry['e_y'][:3]
        ez = hetg.elements[i].geometry['e_z'][:3]
        x = np.array([hetg.hess['xu'][i], hetg.hess['yu'][i], hetg.hess['zu'][i]])
        y = np.array([hetg.hess['xuxf'][i], hetg.hess['yuxf'][i], hetg.hess['zuxf'][i]])
        z = np.array([hetg.hess['xuyf'][i], hetg.hess['yuyf'][i], hetg.hess['zuyf'][i]])
        # all a normalized
        for t in [ex, ey, ez, x, y, z]:
            assert np.allclose(np.linalg.norm(t), 1., atol=1e-4)
        # angle between them is small
        assert np.abs(np.dot(x, ex)) > 0.99
        assert np.abs(np.dot(y, ey)) > 0.99
        assert np.abs(np.dot(z, ez)) > 0.99


def test_groove_dir():
    '''Regression test: Groove direction was not set correctly from input table.'''
    hetg = HETG()
    for i in range(len(hetg.elements)):
        a, c, n = hetg.elements[i].e_groove_coos(np.zeros((1, 2)))
        b = np.array([hetg.hess['xul'][i], hetg.hess['yul'][i], hetg.hess['zul'][i], 0])
        d = np.array([hetg.hess['xud'][i], hetg.hess['yud'][i], hetg.hess['zud'][i], 0])
        # all a normalized
        for x in [a, b, c, d]:
            assert np.allclose(np.linalg.norm(x), 1., atol=1e-4)
        # angle between them is small
        assert np.abs(np.dot(a, b)) > 0.99
        assert np.abs(np.dot(c, d)) > 0.99



# Licensed under GPL version 3 - see LICENSE.rst
import tempfile
import numpy as np
import pytest
import copy
from ...optics import FlatDetector, RectangleAperture
from ...design import RowlandTorus
from ..threejsjson import plot_rays, plot_object, write
from ..utils import MARXSVisualizationWarning

try:
    import jsonschema
    HAS_JSONSCHEMA = True
except ImportError:
    HAS_JSONSCHEMA = False


def test_skip_plot():
    '''Test that plot is skipped with no warning if explicitly
    requested.'''
    det = FlatDetector(zoom=2, position=[2., 3., 4.])
    # First on does not exist, but second one should lead to skip
    # with no warning
    det.display = copy.copy(det.display)
    det.display['shape'] = 'boxxx; None'
    out = plot_object(det)

    # check that only one warning was raised
    assert out is None


def test_box():
    '''Any box shaped element has the same representation so it's
    OK to test just one of them.'''
    det = FlatDetector(zoom=2, position=[2., 3., 4.])
    det.display = det.display
    det.display['opacity'] = 0.1
    out = plot_object(det)
    out_expected = {'geometry': 'BoxGeometry',
                    'geometrypars': (2, 2, 2),
                    'material': 'MeshStandardMaterial',
                    'materialproperties': {'color': '#ffff00',
                                           'opacity': 0.1,
                                           'side': 2,
                                           'transparent': 'true'},
                    'n': 1,
                    'name': "<class 'marxs.optics.detector.FlatDetector'>",
                    'pos4d': [[ 2.,  0.,  0.,  0.,  0.,  2.,  0.,  0.,
                                      0.,  0.,  2.,  0.,  2., 3.,  4.,  1.]]
    }

    # out_expected is only a subset
    for k in out_expected:
        assert out[k] == out_expected[k]


def test_rowland():
    '''Output of Rowland torus'''
    rowland = RowlandTorus(r=2, R=1)
    out = plot_object(rowland)
    out_expected = {'geometry': 'ModifiedTorusBufferGeometry',
                    'material': 'MeshStandardMaterial',
                    'materialproperties': {'color': '#ff4ccc',
                                           'opacity': 0.2,
                                           'side': 2,
                                           'transparent': 'true'},
                    'n': 1,
                    'name': "<class 'marxs.design.rowland.RowlandTorus'>",
    }

    for k in out_expected:
        assert out[k] == out_expected[k]

    assert out['pos4d'] == [np.eye(4).flatten().tolist()]


def test_rays():
    '''Just two rays to make sure the format is right.'''
    rays = plot_rays(np.arange(12).reshape(2,2,3), cmap='jet')
    out_expected = {'color': [[0.0, 0.0, 0.5, 0.0, 0.0, 0.5],
                              [0.0, 0.0, 0.5, 0.0, 0.0, 0.5]],
                    'geometry': 'BufferGeometry',
                    'geometrytype': 'Line',
                    'material': 'LineBasicMaterial',
                    'materialproperties': {'vertexColors': 2},
                    'n': 2,
                    'name': 'Photon list',
                    'pos': [[0, 1, 2, 3, 4, 5], [6, 7, 8, 9, 10, 11]]}

    for k in out_expected:
        assert rays[k] == out_expected[k]

    # See that it validates correctly
    if HAS_JSONSCHEMA:
        with tempfile.TemporaryFile('w') as f:
            write(f, rays)

    else:
        with pytest.warns(UserWarning) as warns:
            with tempfile.TemporaryFile('w') as f:
                write(f, rays)
        assert "without further" in warns[0].message.args[0]


def test_aperture():
    '''Test a plane_with_hole representation.'''
    det = RectangleAperture(zoom=5)
    det.display = det.display
    det.display['opacity'] = 0.3
    out = plot_object(det)
    out_expected = {'faces': [[0, 4, 5, 0, 1, 5, 1, 5, 6, 1, 2, 6, 2, 6, 7, 2, 3, 7,
                               3, 7, 4, 3, 0, 4]],
                    'geometry': 'BufferGeometry',
                    'geometrytype': 'Mesh',
                    'material': 'MeshStandardMaterial',
                    'materialproperties': {'color': '#00bfff',
                                           'opacity': 0.3,
                                           'side': 2,
                                           'transparent': 'true'},
                    'n': 1,
                    'name': "<class 'marxs.optics.aperture.RectangleAperture'>",
                    'pos': [[0.0, 15.0, 15.0, 0.0, -15.0, 15.0, 0.0, -15.0, -15.0,
                             0.0, 15.0, -15.0, 0.0, 5.0, 5.0, 0.0, -5.0, 5.0, 0.0,
                             -5.0, -5.0, 0.0, 5.0, -5.0]]}

    # out_expected is only a subset
    for k in out_expected:
        assert out[k] == out_expected[k]

def test_write():
    '''Write anything to disk.'''
    det = RectangleAperture(zoom=5)
    out = plot_object(det)

    if HAS_JSONSCHEMA:
        with tempfile.TemporaryFile('w') as f:
            write(f, out)

    else:
        with pytest.warns(UserWarning) as warns:
            with tempfile.TemporaryFile('w') as f:
                write(f, out)
        assert "without further" in warns[0].message.args[0]

def test_writelist():
    '''write a list of things to disk.'''
    aper = RectangleAperture(zoom=5)
    det = FlatDetector(zoom=2, position=[2., 3., 4.])
    out = plot_object(aper)
    out1 = plot_object(det)

    if HAS_JSONSCHEMA:
        with tempfile.TemporaryFile('w') as f:
            write(f, [out, out1])

    else:
        with pytest.warns(UserWarning) as warns:
            with tempfile.TemporaryFile('w') as f:
                write(f, [out, out1])
        assert "without further" in warns[0].message.args[0]

def test_write_invalid():
    '''Check that an Exception is raised when trying to write a json that
    is too nested (works only if jsonschema is available).'''
    det = RectangleAperture(zoom=5)
    out = plot_object(det)
    jsonschema = pytest.importorskip("jsonschema")

    with pytest.raises(jsonschema.ValidationError):
        with tempfile.TemporaryFile('w') as f:
            write(f, [[out]])



# Licensed under GPL version 3 - see LICENSE.rst
from __future__ import division

import numpy as np
from astropy.stats import sigma_clipped_stats

from ..optics import FlatDetector, CircularDetector, OrderSelector
from ..design import RowlandTorus
from . import (find_best_detector_position)
from .analysis import sigma_clipped_std
from ..math.geometry import Cylinder


class AnalysisError(Exception):
    pass

def resolvingpower_per_order(gratings, photons, orders, detector=None,
                             colname='det_x'):
    '''Calculate the resolving power in every grating order.

    As input this function takes a Grating Array Structure (``gratings``) and
    a list of photons ready to hit the grating, i.e. the photons have already
    passed through aperture and mirror in e.g. a Chandra-like design. The
    function will take the same input photons and send them through the gas
    looping over the grating orders. In each step of the loop all photons are
    send to the same order, thus the statistical uncertainty on the measured
    spectral resolving power (calculated as ``pos_x / FWHM_x``) is the same
    for every order and is set by the number of photons in the input list.

    As a side effect, the function that selects the grating orders for
    diffraction in ``gas`` will be changed. Pass a deep copy of the GAS if
    this could affect consecutive computations.

    Unlike `~marxs.analysis.gratings.resolvingpower_from_photonlist` this
    function ray-traces the input photon list from the gratings to the
    detectors.

    Parameters
    ----------
    gratings : `marxs.simulator.Parallel`
        Structure that holds individual grating elements,
        e.g. `marxs.design.GratingArrayStructure`
    photons : `astropy.table.Table`
        Photon list to be processed for every order
    orders : np.array of type int
        Order numbers
    detector : marxs optical element or `marxs.design.RowlandTorus` or ``None``.
        Photons are projected onto a detector. There are three ways to define
        this detector:

        - Pass in an instance of an optical element (e.g. a
          `marxs.optics.FlatDetector`).
        - Pass in a `marxs.design.RowlandTorus`. This function will generate
          a detector that follows the Rowland circle.
        - ``None``. A flat detector in the yz plane is used, but the x position
          for this detector is numerically optimized in each step.

    colname : string
        Name of the column that labels the dispersion direction. This is only
        used if a detector instance is passed in explicitly.

    Returns
    -------
    res : np.array
        Resolution defined as FWHM / position for each order.
    fwhm : np.array
        FWHM for each order
    info : dict
        Dictionary with more information, e.g. fit results
    '''
    res = np.zeros_like(orders, dtype=float)
    fwhm = np.zeros_like(orders, dtype=float)
    info = {}

    if detector is None:
        info['method'] = 'Detector position numerically optimized'
        info['fit_results'] = []
        col = 'det_x' # 0 at center, detpix_x is 0 in corner.
        zeropos = 0.
    elif isinstance(detector, RowlandTorus):
        det = CircularDetector(geometry=Cylinder.from_rowland(detector, width=1e5,
                                                              rotation=np.pi,
                                                              kwargs={'phi_lim':[-np.pi/2, np.pi/2]}))
        info['method'] = 'Circular detector on Rowland circle'
        col = 'detpix_x'
        # Find position of order 0.
        # if Rowland torus is tilted, might not be at phi=0.
        pg = photons.copy()
        pg = det(pg)
        # Remove photons that miss a mirror etc.
        pg = pg[pg['probability'] > 0.]
        zeropos, temp1, temp2 = sigma_clipped_stats(pg[col])
    else:
        det = detector
        info['method'] = 'User defined detector'
        col = colname
        # Find position of order 0.
        pg = photons.copy()
        pg = det(pg)
        zeropos, temp1, temp2 = sigma_clipped_stats(pg[col])

    for i, order in enumerate(orders):
        gratingeff = OrderSelector([order])
        gratings.elem_args['order_selector'] = gratingeff
        for elem in gratings.elements:
            elem.order_selector = gratingeff

        pg = photons.copy()
        pg = gratings(pg)
        # Remove photons that slip between the gratings
        pg = pg[(pg['order'] == order) & (pg['probability'] > 0.)]
        if detector is None:
            xbest = find_best_detector_position(pg, objective_func=sigma_clipped_std)
            info['fit_results'].append(xbest)
            det = FlatDetector(position=np.array([xbest.x, 0, 0]), zoom=1e5)
        pg = det(pg)
        meanpos, medianpos, stdpos = sigma_clipped_stats(pg[col])
        fwhm[i] = 2.3548 * stdpos
        res[i] = np.abs((meanpos - zeropos) / fwhm[i])
    return res, fwhm, info


def weighted_per_order(data, orders, energy, gratingeff):
    '''Summarize a per-order table of a quantity such as spectral resolution.

    `marxs.analysis.fwhm_per_order` produces a set of data for each grating
    order, most notably the spectral resolution achieved in every spectral
    order for every energy. In practice, however, most orders see only a very
    small number of photons and will not contribute significantly to the
    observed signal.

    This method provides one way to summarize the data by calculating the
    weighted mean of the resolution for each energy, weighted by the
    probability of photons for be diffracted into that order ( = the expected
    fraction).

    Parameters
    ----------
    data : (N, M) np.array
        Array with data for N orders and M energies, e.g. spectral resolution
    orders : np.array of length N of type int
        Order numbers
    energy : np.array for length M of type float
        Energies for each entry in resolution
    gratingeff : `marxs.optics.gratings.EfficiencyFile`
        Object that holds the probability that a photon on a certain energy is
        diffracted to a specific order.

    Returns
    -------
    weighted_res : np.array of length M
        Weighted resolution for each energy
    '''
    if len(orders) != data.shape[0]:
        raise ValueError('First dimension of "data" must match length of "orders".')
    if len(energy) != data.shape[1]:
        raise ValueError('Second dimension of "data" must match length of "energy".')

    weights = np.zeros_like(data)
    for i, o in enumerate(orders):
        ind_o = (gratingeff.orders == o).nonzero()[0]
        if len(ind_o) != 1:
            raise KeyError('No data for order {0} in gratingeff'.format(o))
        en_sort = np.argsort(gratingeff.energy)
        weights[i, :] = np.interp(energy, gratingeff.energy[en_sort],
                                  gratingeff.prob[:, ind_o[0]][en_sort])

    return np.ma.average(data, axis=0, weights=weights)


def resolvingpower_from_photonlist(photons, orders,
                                   col='proj_x', zeropos=None,
                                   ordercol='order'):
    '''Calculate the resolving power for several grating orders

    If fewer than 20 photons are detected in a single order, this function
    returns nan values.

    Unlike `~marxs.analysis.gratings.resolvingpower_per_order` this method does
    not run any ray-trace simulations. It just extracts information from a
    photon list that is passed in.

    Parameters
    ----------
    photons : `astropy.table.Table`
        Photon event list
    orders : np.array
        Orders for which the resolving power will be calculated
    col : string
        Column name for the column holding the dispersion coordinate.
    zeropos : float or ``None``
        Value of column `col` where the zeroth order is found. If not given,
        this is calculated (assuming the zeroth order photons are part of the
        event list).
    ordercol : string
        Name of column that lists grating order for each photon

    Returns
    -------
    res : np.array
        resolving power for each order
    pos : np.array
        mean value of ``col`` for each order
    std : np.array
        standard deviation of the distribution of ``col`` for each order
    '''
    if zeropos is None:
        ind = (photons[ordercol] == 0)
        if ind.sum() < 20:
            raise AnalysisError('Too few photons in list to determine position of order 0 automatically.')
        zeropos, medzeropos, stdzero = sigma_clipped_stats(photons[col][ind])

    pos = np.zeros_like(orders, dtype=float)
    std = np.zeros_like(orders, dtype=float)

    for i, o in enumerate(orders):
        ind = (photons[ordercol] == o)
        if ind.sum() > 20:
            meanpos, medianpos, stdpos = sigma_clipped_stats(photons[col][ind])
        else:
            meanpos, stdpos = np.nan, np.nan
        pos[i] = meanpos
        std[i] = stdpos
    res = np.abs(pos - zeropos) / (std * 2.3548)
    return res, pos, std


def effectivearea_from_photonlist(photons, orders, n_photons, A_geom=1.,
                                   ordercol='order'):
    '''Calculate the effective area several grating orders

    This is based on the probabilities of the photons in the list, so
    this photons list must already account for all instrument
    components, for example, do not forgot to set the probability to 0
    for photons that falls into a chip gap.

    Parameters
    ----------
    photons : `astropy.table.Table`
        Photon event list
    orders : np.array
        Orders for which the resolving power will be calculated
    n_photons : int
        Number of photons originally simulated
    A_geom : number
        Geometric area of aperture that was used for photon list.
    ordercol : string
        Name of column that lists grating order for each photon

    Returns
    -------
    aeff : np.array
        effective area for each order
    '''
    aeff = np.zeros(len(orders))
    for i, o in enumerate(orders):
        aeff[i] = photons['probability'][photons[ordercol] == o].sum()
    return aeff / n_photons * A_geom



# Licensed under GPL version 3 - see LICENSE.rst
from __future__ import division

import numpy as np
import scipy.optimize
from astropy.stats import sigma_clipped_stats

from ..optics import FlatDetector

def sigma_clipped_std(photons, colname='det_x', **kwargs):
    '''Return stddev of sigma-clipped data.

    Parameters
    ----------
    photons : `astropy.table.Table`
        Input photon list.
    colname : string
        Column name of the photon distribution to be minimized.
        The default is set for detectors that look for a grating signal
        (which is dispersed in ``det_y`` direction).
    kwargs : see `astropy.stats.sigma_clipped_stats`
        All keyword arguments are passed to `astropy.stats.sigma_clipped_stats`.
    '''
    mean, median, std = sigma_clipped_stats(photons[colname].data, **kwargs)
    return std


def mean_width_2d(photons):
    '''Return average distance from center of det_x, det_y distribution.

    Parameters
    ----------
    photons : `astropy.table.Table`
        Input photon list.
    '''
    x = photons['det_x'].data
    y = photons['det_y'].data
    r = np.linalg.norm(np.stack([x - x.mean(), y - y.mean()]), axis=0)
    return np.sum(r) / len(r)



def find_best_detector_position(photons,
                                objective_func=sigma_clipped_std,
                                objective_func_args={'colname': 'det_x'},
                                orientation=np.eye(3), **kwargs):
    '''Numerically find the position of best focus.

    This routine places detectors at different positions and
    calculates the width of the photon distribution for each position.
    The detector is moved perpendicular to its plane.
    As a side effect, ``photons`` will be set to the intersection with
    the last try of the detector position.

    Parameters
    ----------
    photons : `astropy.table.Table`
        Input photon list.
    objective_func : function
        Function that accepts a np.array as input and return the width.
    object_func_args : dict
        Dict of any other options for the `objective_func`.
    rotation : np.array of shape (3,3)
        Rotation matrix for the detector. By default the detector is parallel
        to the yz plane of the global coordinate system (see `pos4d`).
    kwargs : see `scipy.optimize.minimize`
        All other keyword argument will be passed to `scipy.optimize.minimize`.
    Returns
    -------
    opt : OptimizeResult
        see `scipy.optimize.minimize_scalar`

    '''
    def width(x, photons):
        mdet = FlatDetector(position=np.dot(orientation, np.array([x, 0, 0])),
                            orientation=orientation,
                            zoom=1e5, pixsize=1.)
        p = mdet(photons.copy())
        return objective_func(p, **objective_func_args)

    return scipy.optimize.minimize_scalar(width, args=(photons,), options={'maxiter': 20, 'disp': False},
                                   **kwargs)


def detected_fraction(photons, labels, col='order'):
    '''Calculate the fraction of photons detected for some integer label

    While written for calculating Aeff per order, this can be used with any discrete
    quantity, e.g. Aeff per CCD.

    Parameters
    ----------
    photons : `astropy.table.Table`
        Photon event list
    labels : np.array
        Numeric (integer) labels that are found in column ``col``. When, e.g.,
        the effective area per order is calculated, this array should contain
        the order numbers for which the calculation will be done.
    col : string
        Column name for the order column.

    Returns
    -------
    prop : np.array
        Probability for a photon in a specific order to be detected.
    '''
    prob = np.zeros_like(labels, dtype=float)
    for i, o in enumerate(labels):
        ind = (photons[col] == o)
        if filterfunc is not None:
            ind = ind & filterfunc(photons)
        prob[i] = np.sum(photons['probability'][ind]) / len(photons)
    return prob



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import print_function


import glob
import os
import sys

import ah_bootstrap
from setuptools import setup

#A dirty hack to get around some early import/configurations ambiguities
if sys.version_info[0] >= 3:
    import builtins
else:
    import __builtin__ as builtins
builtins._ASTROPY_SETUP_ = True

from astropy_helpers.setup_helpers import (register_commands, get_debug_option,
                                           get_package_info)
from astropy_helpers.git_helpers import get_git_devstr
from astropy_helpers.version_helpers import generate_version_py

# Get some values from the setup.cfg
try:
    from ConfigParser import ConfigParser, NoOptionError
except ImportError:
    from configparser import ConfigParser, NoOptionError
conf = ConfigParser()

conf.read(['setup.cfg'])
metadata = dict(conf.items('metadata'))

PACKAGENAME = metadata.get('name', 'packagename')
DESCRIPTION = metadata.get('description', 'Astropy affiliated package')
AUTHOR = metadata.get('author', '')
AUTHOR_EMAIL = metadata.get('author_email', '')
LICENSE = metadata.get('license', 'unknown')
URL = metadata.get('url', 'http://astropy.org')

# Get the long description from the package's docstring
__import__(PACKAGENAME)
package = sys.modules[PACKAGENAME]
LONG_DESCRIPTION = package.__doc__

# Store the package name in a built-in variable so it's easy
# to get from other parts of the setup infrastructure
builtins._ASTROPY_PACKAGE_NAME_ = PACKAGENAME

# Populate the dict of setup command overrides; this should be done before
# invoking any other functionality from distutils since it can potentially
# modify distutils' behavior.
cmdclassd = register_commands()

# Freeze build information in version.py
generate_version_py()

# Treat everything in scripts except README.rst as a script to be installed
scripts = [fname for fname in glob.glob(os.path.join('scripts', '*'))
           if os.path.basename(fname) != 'README.rst']


# Get configuration information from all of the various subpackages.
# See the docstring for setup_helpers.update_package_files for more
# details.
package_info = get_package_info()

# Add the project-global data
package_info['package_data'].setdefault(PACKAGENAME, [])
package_info['package_data'][PACKAGENAME].append('data/*')

# Define entry points for command-line scripts
entry_points = {'console_scripts': []}

entry_point_list = conf.items('entry_points')
for entry_point in entry_point_list:
    entry_points['console_scripts'].append('{0} = {1}'.format(entry_point[0],
                                                              entry_point[1]))

# Include all .c files, recursively, including those generated by
# Cython, since we can not do this in MANIFEST.in with a "dynamic"
# directory name.
c_files = []
for root, dirs, files in os.walk(PACKAGENAME):
    for filename in files:
        if filename.endswith('.c'):
            c_files.append(
                os.path.join(
                    os.path.relpath(root, PACKAGENAME), filename))
package_info['package_data'][PACKAGENAME].extend(c_files)

# Note that requires and provides should not be included in the call to
# ``setup``, since these are now deprecated. See this link for more details:
# https://groups.google.com/forum/#!topic/astropy-dev/urYO8ckB2uM

setup_args = {'name': PACKAGENAME,
              'description': DESCRIPTION,
              'scripts': scripts,
              'install_requires': ['astropy', 'transforms3d'],
              'author': AUTHOR,
              'author_email': AUTHOR_EMAIL,
              'license': LICENSE,
              'url': URL,
              'long_description': LONG_DESCRIPTION,
              'cmdclass': cmdclassd,
              'zip_safe': False,
              'use_2to3': False,
              'entry_points': entry_points,
              'setup_requires': ['astropy', 'transforms3d'],  # We use six from astropy.extern
              }
setup_args.update(package_info)

# check is MARX C code is configured in setup.py and if it is, add to
# appropriate arguments to setup args
try:
    marxscr = conf.get('MARX', 'srcdir')
    marxlib = conf.get('MARX', 'libdir')
    if marxscr and marxlib:
        setup_args['cffi_modules'] = ["marxs/optics/marx_build.py:ffi"]
        setup_args['install_requires'].append("cffi>=1.0.0")
        setup_args['setup_requires'].append("cffi>=1.0.0")
        print('Using MARX C code at {0}\n and compiled libraries at {1}'.format(marxscr, marxlib))
    else:
        print('MARX C code is not configured in setup.cfg - modules will be disabled.')
except NoOptionError:
    print('MARX C code is not configured in setup.cfg - modules will be disabled.')

setup(**setup_args)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from astropy.table import Table
import transforms3d

from ..detector import FlatDetector, CircularDetector
from ...tests import closeornan
from ...math.utils import h2e
from ...design import RowlandTorus
from ...math.geometry import Cylinder

def test_pixelnumbers():
    pos = np.array([[0, 0., -0.25, 1.],
                    [0., 9.9, 1., 1.],
                    [0., 10.1, 1., 1.]])
    dir = np.ones((3,4), dtype=float)
    dir[:, 3] = 0.
    photons = Table({'pos': pos, 'dir': dir,
                     'energy': [1,2., 3.], 'polarization': [1.,2, 3.], 'probability': [1., 1., 1.]})
    det = FlatDetector(zoom=[1., 10., 5.], pixsize=0.5)
    assert det.npix == [40, 20]
    assert det.centerpix == [19.5, 9.5]
    photons = det(photons)
    assert closeornan(photons['det_x'], np.array([0, 9.9, np.nan]))
    assert closeornan(photons['det_y'], np.array([-0.25, 1., np.nan]))
    assert closeornan(photons['detpix_x'], np.array([19.5, 39.3, np.nan]))
    assert closeornan(photons['detpix_y'], np.array([9, 11.5, np.nan]))

    # Regression test: This was rounded down to [39, 39] at some point.
    det = FlatDetector(pixsize=0.05)
    assert det.npix == [40, 40]


def test_nonintegerwarning(recwarn):
    det = FlatDetector(zoom=np.array([1.,2.,3.]), pixsize=0.3)
    w = recwarn.pop()
    assert 'is not an integer multiple' in str(w.message)


def test_CircularDetector_from_Rowland():
    '''If a circular detector is very narrow, then all points should be
    very close to the Rowland torus.'''
    rowland = RowlandTorus(R=6e4, r=5e4, position=[123., 345., -678.],
                           orientation=transforms3d.euler.euler2mat(1, 2, 3, 'syxz'))
    detcirc = CircularDetector(geometry=Cylinder.from_rowland(rowland, width=1e-6))
    phi = np.mgrid[0:2.*np.pi:360j]
    points = detcirc.geometry.parametric_surface(phi)
    # Quartic < 1e5 is very close for these large values of r and R.
    assert np.max(np.abs(rowland.quartic(h2e(points)))) < 1e5



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from astropy.table import Column
import astropy.units as u
import transforms3d

from ..optics.base import FlatOpticalElement
from .source import Source
from ..math.polarization import polarization_vectors
from ..math.utils import h2e, e2h


class FarLabPointSource(Source, FlatOpticalElement):
    '''Simple in-lab point source used with aperture

    - assumes point source is far from a rectangular aperture, and only the photons that pass through are tracked
    - photon start positions uniformly distributed within rectangular aperture (reasonable approximation if source is far)
    - photon direction determined by location of source, and selected photon starting position
    - aperture is in its local y-z plane

    Parameters
    ----------
    sourcePos: 3 element list
        3D coordinates of photon source (not aperture)
    kwargs: ``pos4d`` or ``position``, ``orientation``, and ``zoom`` can be used to set the position,
        size and orientation of the rectangular apeture; see `pos4d` for details.
        Other keyword arguments include ``flux``, ``energy`` and ``polarization``.
        See `Source` for details.
    '''
    def __init__(self, sourcePos, **kwargs):
        self.sourcePos = sourcePos
        if not 'flux' in kwargs:
            kwargs['flux'] = 1 / u.s
        super(FarLabPointSource, self).__init__(geomarea=None, **kwargs)

    @u.quantity_input
    def generate_photons(self, exposuretime: u.s):
        photons = super(FarLabPointSource, self).generate_photons(exposuretime)
        n = len(photons)
        # randomly choose direction - photons uniformly distributed over aperture area
        # measurements in mm
        pos = np.dot(self.pos4d, np.array([np.zeros(n),
                                           np.random.uniform(-1, 1, n),
                                           np.random.uniform(-1, 1, n),
                                           np.ones(n)]))

        dir = np.array([pos[0, :] - self.sourcePos[0],
                        pos[1, :] - self.sourcePos[1],
                        pos[2, :] - self.sourcePos[2],
                        np.zeros(n)])
        photons['pos'] = pos.T
        photons['dir'] = dir.T
        photons['polarization'] = polarization_vectors(dir.T, photons['polangle'])
        return photons


class LabPointSourceCone(Source):
    '''In-lab point source

    - Photons are uniformly distributed in all directions in cone (from the point). Cone is meant to refer to the volume swept out by a solid angle in a unit sphere centered at the point source.
    - Photon start position is source position (tip of cone)

    Parameters
    ----------
    position: array-like of shape(3,)
        Eukledian coordinates of photon source. Default is at the origin
        of the coordinate system.
    half_opening: float
        This is half the openning angle of the cone. It is given in steradians.
        The default is pi, which distributes the photons evenly over the entire
        sphere.
    direction: array-like of shape (3,)
        This is the direction of the center of the cone.
        Default is the positive x-axis.
    kwargs : see `Source`
        Other keyword arguments include ``flux``, ``energy`` and ``polarization``.
        See `Source` for details.
    '''
    def __init__(self, position=[0, 0, 0], half_opening=np.pi,
                 direction=[1., 0., 0.],
                 **kwargs):
        if (len(position) != 3) or (len(direction) != 3):
            raise ValueError('Direction and position are expected in Eukledian coordinates.')
        self.dir = e2h(np.asanyarray(direction) / np.linalg.norm(direction), 0)
        self.position = e2h(np.asanyarray(position), 1)
        self.half_opening = half_opening
        if not 'flux' in kwargs:
            kwargs['flux'] = 1 / u.s
        super(LabPointSourceCone, self).__init__(geomarea=None, **kwargs)

    @u.quantity_input
    def generate_photons(self, exposuretime: u.s):
        photons = super(LabPointSourceCone, self).generate_photons(exposuretime)
        n = len(photons)

        # assign position to photons
        pos = np.ones((4, n)) * self.position[:, None]

        # Randomly choose direction - photons directions randomly distributed inside cone.
        # Visualize in arbitrary x-y-z coordinate system (to be reconciled with class parameters later)
        # Angle from pole (z-axis) = phi. Angle from x-axis on x-y plane is theta
        # This cone is temporarily centered about the z axis.
        theta = np.random.uniform(0, 2 * np.pi, n)
        #this is the fractional surface area swept out by delta
        fractionalArea = 2 * np.pi * (1 - np.cos(self.half_opening)) / (4 * np.pi)
        v = np.random.uniform(0, fractionalArea, n)
        phi = np.arccos(1 - 2 * v)
        # For computation of phi see http://www.bogotobogo.com/Algorithms/uniform_distribution_sphere.php

        dir = np.array([np.cos(theta) * np.sin(phi),
                        np.sin(theta) * np.sin(phi),
                        np.cos(phi),
                        np.zeros(n)])

        # Now we have all directions for n photons in dir.
        # Now we rotate dir to align with the given direction: self.dir
        # To find axis of rotation: cross self.dir with z
        axis = np.cross(h2e(self.dir), [0, 0, 1])

        angle = np.arccos(self.dir[2]) # Simplified

        rotationMatrix = transforms3d.axangles.axangle2aff(axis, -angle)

        # The aligned directions are:
        dir = np.dot(rotationMatrix, dir)


        photons.add_column(Column(name='pos', data=pos.T))
        photons.add_column(Column(name='dir', data=dir.T))
        photons.add_column(Column(name='polarization', data=polarization_vectors(dir.T, photons['polangle'])))
        return photons



# Licensed under GPL version 3 - see LICENSE.rst
'''This module contains tests that are done for all opticial elements as
opposed to tests that apply to one class of elements only (e.g. gratings).
Tests here include simple things that guarantee that all elements comply with
the MARXS interface (e.g. each element must return a name and description) and
more complex consistency checks that are required for every type of optical
element (e.g. the polarization vector must be perpendicular to the photon
propagation direction in each and every case).
'''
from collections import OrderedDict
import copy

import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy import units as u
import transforms3d.axangles

from .. import (RectangleAperture, ThinLens, FlatDetector, CircularDetector,
                FlatGrating, OrderSelector,
                MarxMirror, CircleAperture, MultiAperture)

from ..aperture import BaseAperture
from ...source import PointSource, FixedPointing
from ..base import FlatStack
from ...base import _parse_position_keywords
from ...design import (RowlandTorus, GratingArrayStructure,
                       LinearCCDArray, RowlandCircleArray,
                       RectangularGrid)
from ..baffle import Baffle
from ..multiLayerMirror import MultiLayerMirror
from ...simulator import Sequence
from ...missions.chandra.hess import HETG
from ...missions.mitsnl.catgrating import (QualityFactor, L1,
                                           L2Abs, L2Diffraction,
                                           CATL1L2Stack,
                                           NonParallelCATGrating)
from ..scatter import RadialMirrorScatter, RandomGaussianScatter
from ..filter import EnergyFilter


# Initialize all optical elements to be tested
mytorus = RowlandTorus(0.5, 0.5)

all_oe = [ThinLens(focallength=100),
          RectangleAperture(),
          CircleAperture(),
          MultiAperture(elements=[RectangleAperture(),
                                  CircleAperture(position=[0, 200, 0])]),
          FlatDetector(pixsize=2., zoom=100.),
          CircularDetector(),
          FlatGrating(d=0.001, order_selector=OrderSelector([0])),
          MarxMirror(parfile='marxs/optics/hrma.par'),
          GratingArrayStructure(mytorus, d_element=0.1, x_range=[0.5, 1.], radius=[0,.5],
                                elem_class=FlatGrating,
                                elem_args={'zoom':0.05, 'd':0.002,
                                           'order_selector': OrderSelector([1])
                                           }),
          LinearCCDArray(mytorus, d_element=0.05, x_range=[0., 0.5],
                          radius=[0., 0.5], phi=0., elem_class=FlatGrating,
                         elem_args={'zoom': 0.05, 'd':0.002,
                                    'order_selector': OrderSelector([2])
                                }),
          RowlandCircleArray(rowland=mytorus, elem_class=FlatGrating,
                             elem_args={'zoom': 0.04, 'd': 1e-3,
                                        'order_selector': OrderSelector([2])},
                             d_element=0.1,theta=[np.pi - 0.2, np.pi + 0.1]),
          RectangularGrid(rowland=mytorus, elem_class=FlatGrating,
                          elem_args={'zoom': 0.04, 'd': 1e-3,
                                     'order_selector': OrderSelector([-1, 0])},
                          d_element=0.1, x_range=[0.8, 1.1], y_range=[0, 0.1],
                          z_range=[0, 0.1]),
          Baffle(),
          # MLMirror assumes 45 deg angle and will return nan otherwise
          MultiLayerMirror(reflFile='./marxs/optics/data/testFile_mirror.txt',
                           testedPolarization='./marxs/optics/data/ALSpolarization2.txt',
                           orientation=transforms3d.axangles.axangle2mat([0,1,0], np.pi/4)),
          Sequence(elements=[]),
          HETG(),
          RadialMirrorScatter(inplanescatter=0.1 * u.arcsec),
          RandomGaussianScatter(scatter=.1 * u.arcmin),
          # not a useful filterfunc, but OK for testing with a few other dependencies
          EnergyFilter(filterfunc=lambda x: np.abs(np.cos(x))),
          FlatStack(elements=[EnergyFilter, FlatDetector], keywords=[{'filterfunc': lambda x: 0.5}]),
          # elements in MIT SNL
          # Most of them have defaults for all parameters
          QualityFactor(),
          L1(order_selector=OrderSelector([0])),
          L2Abs(),
          L2Diffraction(),
          CATL1L2Stack(order_selector=OrderSelector([2])),
          NonParallelCATGrating(order_selector=OrderSelector([2]), d=0.0001),
          ]

# Each elements will be used multiple times.
# Can I add a test to check that using them does not leave any
# extra attributes etc., but that they come out in a clean state?
# How would I do that?


# Make a test photon list
# Some of this should be separate tests, e.g. source position vs. pointing.
# Can I vary energy for e.g. grating?
mysource = PointSource(coords=SkyCoord(30., 30., unit='deg'))
masterphotons = mysource.generate_photons(500 * u.s)
mypointing = FixedPointing(coords=SkyCoord(30., 30., unit='deg'))
masterphotons = mypointing(masterphotons)
myslit = RectangleAperture()
masterphotons = myslit(masterphotons)
masterphotons['order'] = 0


@pytest.fixture(autouse=True)
def photons():
    '''make a photons table - fresh for every test'''
    photons = masterphotons.copy()
    assert id(photons['dir']) != id(masterphotons['dir'])
    assert np.all(photons['dir'] == photons['dir'])
    return photons



mark = pytest.mark.parametrize


@mark('elem', all_oe)
class TestOpticalElementInterface:
    '''Test that all optical elements follow the interface convention.

    At the same time, these tests guarantee that the modules work at all
    (e.g. the can be initialized etc.). Not as good as checking for correctness
    (that has to happen in individual tests), but it's a start.
    '''
    def test_one_vs_many_in_call_signature(self, photons, elem):
        '''processing a single photon should give same result as photon list'''
        assert np.all(photons['dir'] == photons['dir'])
        p = photons[[0]]
        # Process individually
        single = photons[0]
        if hasattr(elem, 'process_photon'):
            try:
                # For apertures input pos is ignored, but still needs to be there
                # to keep the function signature consistent.
                dir, pos, energy, pol, prob = elem.process_photon(single['dir'], single['pos'], single['energy'], single['polarization'])
                compare_results = True
            except NotImplementedError:
                # It's OK if only the a vectorized process_photons is implemented
                # but if both exist, they need to return the same answer.
                compare_results = False
        else:
            compare_results = False
        # Process as table
        if isinstance(elem, BaseAperture):
            photons.remove_column('pos')
        p = elem(photons)

        if compare_results:
            # We can only compare, if one and many did run.
            assert np.all(dir == p['dir'][0])
            assert np.all(pos == p['pos'][0])
            assert energy == p['energy'][0]
            assert (np.all(pol == p['polarization'][0]) or
                    (np.all(np.isnan(pol)) and np.all(np.isnan(p['polarization'][0]))))
            assert prob == p['probability'][0]

    def test_desc(self, elem):
        '''every elements should return a Ordered Dict for description'''
        des = elem.describe()
        assert isinstance(des, OrderedDict)
        assert len(des) > 0

    def test_prob(self, photons, elem):
        '''For every type of element, the value for 'probability' can only go down.'''
        if isinstance(elem, BaseAperture):
            photons.remove_column('pos')

        photons['probability'] = 1e-4
        photons = elem(photons)
        assert np.all(photons['probability'] <= 1e-4)

    def test_polarization_direction(self, photons, elem):
        '''For every type of element, the polarization vector must still be
        perpendicular to the direction of the photon after it passed through
        the element. Most elements do not change the direction of the photon,
        so this test is not really relevant, but testing the full list of
        elements here is the only way in make sure that this test is not
        forgotten when individual elements are implemented.'''
        if isinstance(elem, BaseAperture):
            photons.remove_column('pos')

        photons = elem(photons)
        assert np.allclose(np.einsum('...i,...i', photons['dir'].data,
                                     photons['polarization'].data), 0.)
        # Also, the polarization vector must still be a unit vector.
        assert np.allclose(np.linalg.norm(photons['polarization'], axis=1), 1)


    def test_metadata_evolution(self, photons, elem):
        '''In almost all cases, the metadata passed into an element should
        be returned unchanged. New keys can be added all the time, but it is
        extremely rare that an element changes an existing meta-data value.
        (Those cases need to be marked xfail by hand here or be listed in an
        exclusion list when they come up.)

        Regression test for #88.
        '''
        meta_in = copy.deepcopy(photons.meta)
        if isinstance(elem, BaseAperture):
            photons.remove_column('pos')
        p = elem(photons)
        assert all(item in p.meta.items() for item in meta_in.items())


def test_parse_position_keywords_zoom_dimension():
    '''test proper error messages for zoom keyword'''
    for zoom in [np.ones(1), np.ones(2), np.ones(4)]:
        with pytest.raises(ValueError) as e:
            pos4d = _parse_position_keywords({'zoom': zoom})
            assert 'must has three elements' in str(e.value)



# Licensed under GPL version 3 - see LICENSE.rst
def get_package_data():
    # I could not figure out how to get package_data
    # to deal with a directory hierarchy of files, so just explicitly list.
    return {
        'marxs.missions.chandra': ['HESSdesign.rdb',
                                  'tests/sim_asol.fits',
                         ]
    }



# Licensed under GPL version 3 - see LICENSE.rst
from .rowland import (RowlandTorus,
                      GratingArrayStructure, LinearCCDArray,
                      RectangularGrid,
                      RowlandCircleArray,
                      find_radius_of_photon_shell)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import astropy.units as u

from marxs.source.labSource import LabPointSourceCone
from marxs.optics.multiLayerMirror import MultiLayerMirror
from marxs.optics.detector import FlatDetector

def test_src_mir_det():
    '''This tests that source, multilayer mirror, and detector run without producing any errors.
    '''
    string1 = 'marxs/optics/data/A12113.txt'
    string2 = 'marxs/optics/data/ALSpolarization2.txt'
    a = 2**(-0.5)
    rotation1 = np.array([[a, 0, -a],
                          [0, 1, 0],
                          [a, 0, a]])
    rotation2 = np.array([[0, 0, 1],
                          [0, 1, 0],
                          [-1, 0, 0]])

    source = LabPointSourceCone([10., 0., 0.], flux=100. / u.s)
    mirror = MultiLayerMirror(reflFile=string1, testedPolarization=string2,
                              position=np.array([0., 0., 0.]),
                              orientation=rotation1)
    detector = FlatDetector(1., position=np.array([0., 0., 10.]),
                            orientation=rotation2, zoom = np.array([1, 100, 100]))

    photons = source.generate_photons(100 * u.s)
    photons = mirror(photons)
    photons = detector(photons)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import pytest
from astropy.coordinates import SkyCoord
import astropy.units as u
from ... import source, optics
from ...math.utils import h2e


@pytest.mark.parametrize("ra", [0., 30., -60.])
def test_PerfectLens(ra):
    mysource = source.PointSource(coords=SkyCoord(0., ra, unit="deg"))
    mypointing = source.FixedPointing(coords=SkyCoord(0., 0., unit='deg'))
    myslit = optics.RectangleAperture(zoom=2, position=[+100, 0, 0])
    f = 1000
    lens = optics.PerfectLens(focallength=f, zoom=400)

    photons = mysource.generate_photons(11 * u.s)
    photons = mypointing(photons)
    photons = myslit(photons)
    assert np.allclose(h2e(photons['pos'].data)[:, 0], 100.)

    photons = lens(photons)
    assert np.allclose(h2e(photons['pos'].data)[:, 0], 0.)

    # How far away do I need to put a detector to hit the focal point?
    d = f * np.cos(np.deg2rad(ra))
    mdet = optics.FlatDetector(pixsize=0.01, position=np.array([-d, 0, 0]), zoom=1e5)
    photons = mdet(photons)
    assert np.std(photons['det_x']) < 1e-4
    assert np.std(photons['det_y']) < 1e-4

    # Check that they actually diverge for other detector placement
    mdet = optics.FlatDetector(pixsize=0.01, position=np.array([-2 * d, 0, 0]), zoom=1e5)
    photons = mdet(photons)
    assert np.std(photons['det_x']) > 1e-4
    assert np.std(photons['det_y']) > 1e-4



# Licensed under GPL version 3 - see LICENSE.rst
from copy import deepcopy
import numpy as np

from ..rowland import RowlandTorus, GratingArrayStructure
from ..uncertainties import generate_facet_uncertainty
from ...optics import FlatGrating, OrderSelector

def test_uncertainty_generation():
    '''The best way to test that the output format is reasonable is to use it.

    I do not really have an idea how I test that the distribution is right without
    doing the exact reverse (fit the distribution) of the procedure we are
    testing here.
    Also, generating the GAS is a fairly expensive operation, so I do not want to
    make too many elements unless there is a good reason for it.
    '''

    np.random.seed(10)

    rowland  = RowlandTorus(5e3, 5e3)
    gas = GratingArrayStructure(rowland, d_element=2., x_range=[9e3, 1.1e4],
                                radius=[2, 15], phi=[0, np.pi],
                                elem_class=FlatGrating,
                                elem_args={'d': 2e-4,
                                           'order_selector': OrderSelector([1])},
    )
    oldgaspos = deepcopy(gas.elem_pos)
    for a, b in zip(oldgaspos, gas.elements):
        # Check position record in GAS and in elements is in snyc
        assert np.allclose(a, b.pos4d)

    # Note that input is list, which will be converted to ndarray internally
    gas.elem_uncertainty = generate_facet_uncertainty(len(gas.elem_uncertainty),
                                                      [0, 0, 1.], np.zeros(3))
    # Now apply the uncertainties to make sure it works
    gas.generate_elements()

    for a, b in zip(gas.elem_pos, gas.elements):
        # Check position record in GAS and in elements is still in snyc
        assert ~np.allclose(a, b.pos4d)

    # Check that the shifts are right
    shifts = np.array([e[2, 3] - f.pos4d[2, 3] for e, f in zip(oldgaspos, gas.elements)])
    assert np.isclose(np.mean(shifts), 0., atol=0.1)
    assert np.isclose(np.std(shifts), 1., atol=0.1)



# Licensed under GPL version 3 - see LICENSE.rst
from os import path
from configparser import ConfigParser
from cffi import FFI

with open ("marxs/optics/cdef.txt", "r") as myfile:
    cdeftxt=myfile.read()

ffi = FFI()
ffi.cdef(cdeftxt)


conf = ConfigParser()
# When setup.py is run, then setup.cfg is in the current directory
# However, when runngin this in pytest, it's in ../
# So offer both options here (files not found are silently ignored).
conf.read(['setup.cfg', '../setup.cfg'])
marxscr = conf.get('MARX', 'srcdir')
marxlib = conf.get('MARX', 'libdir')

sources = [('pfile', 'src', 'pfile.c'), ('marx', 'libsrc', 'mirror.c')]
headers = [('pfile', 'src'), ('jdmath', 'src') , ('jdfits', 'src'),
           ('marx', 'src',), ('marx', 'libsrc',), ('src',), ('libsrc',)]

ffi.set_source("_marx",
'''
# include "pfile.h"
# include "_pfile.h"

# include <jdmath.h>
# include "marx.h"
''',
                        sources=[path.join(marxscr, *f) for f in sources],
                        include_dirs=[path.join(marxscr, *f) for f in headers],
                        libraries=['marx', 'pfile', 'jdmath', 'jdfits'],
                        library_dirs=[marxlib]
                        )

if __name__ == "__main__":
    ffi.compile()
